# Copyright (c) 2022, Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""ONNX model execution utilities for FINN.

This module provides enhanced ONNX model execution capabilities that extend
the base QONNX execution functionality with FINN-specific features, including
RTL simulation support and model debugging utilities.

Key Functions:
- execute_onnx: Enhanced ONNX execution with RTL simulation support
- execute_onnx_and_make_model: Create debug models with intermediate activations
- compare_execution: Compare outputs between two models
"""

import copy
import numpy as np
import qonnx.analysis.topology as ta
from qonnx.core.onnx_exec import execute_onnx as execute_onnx_base

from finn.core.rtlsim_exec import rtlsim_exec


def execute_onnx(model, input_dict, return_full_exec_context=False, start_node=None, end_node=None):
    """Executes given ONNX ModelWrapper with given named inputs.
    If return_full_exec_context is False, a dict of named outputs is returned
    as indicated by the model.graph.output.
    If return return_full_exec_context is True, the full set of tensors used by
    the execution (including inputs, weights, activations and final outputs)
    will be returned as a dict.
    When start_node and end_node are set to None, the whole graph is executed.
    If they are set to particular ONNX nodes, only the subgraph between (and
    including) those nodes is executed.
    """

    # check if model has an execution mode set
    # if None, execute model node using the QONNX-provided execute_onnx impl
    # if set to "rtlsim" execute model using xsi
    model_exec_mode = model.get_metadata_prop("exec_mode")
    if (model_exec_mode is None) or (model_exec_mode == ""):
        return execute_onnx_base(model, input_dict, return_full_exec_context, start_node, end_node)
    elif model_exec_mode == "rtlsim":
        # check sanity of model and then use stitched IP for rtlsim
        if not model.check_all_tensor_shapes_specified():
            raise Exception("Found unspecified tensor shapes, try infer_shapes")
        ret = model.analysis(ta.nodes_topologically_sorted)
        assert (
            ret["nodes_topologically_sorted"] is True
        ), """Nodes must be
        topologically sorted."""

        graph = model.graph
        # first, we need to make sure that every variable required by the graph has
        # some buffer associated with it. this includes graph inputs (which includes
        # the input data as well as the trained parameters) and the graph ValueInfo
        # (intermediate tensors between layers)
        # this is provided by the execution_context, which is a dict of np.ndarray
        execution_context = model.make_empty_exec_context()
        # fill in any inputs provided to this function
        for inp_name in input_dict.keys():
            if inp_name in execution_context:
                if execution_context[inp_name].shape == input_dict[inp_name].shape:
                    execution_context[inp_name] = input_dict[inp_name]
                else:
                    raise Exception(
                        "Shape mismatch for provided input %s: found %s expected %s "
                        % (
                            inp_name,
                            str(execution_context[inp_name].shape),
                            str(input_dict[inp_name].shape),
                        )
                    )

        # use stitched IP for rtlsim
        rtlsim_exec(model, execution_context)
    else:
        raise Exception(
            """Metadata property "exec_mode" is set to an unknown value. Can be left
            unset or has to be set to "rtlsim" for execution using xsi!"""
        )

    if return_full_exec_context:
        return execution_context
    else:
        # provide outputs as dict
        output_dict = dict()
        for out_tensor in graph.output:
            out_name = out_tensor.name
            output_dict[out_name] = execution_context[out_name]
        return output_dict


def execute_onnx_and_make_model(model, input_dict):
    """Executes given ONNX ModelWrapper with given named inputs and return a new
    ModelWrapper where an initializer is provided for each tensor as taken from
    the execution. This new model is useful for debugging, since it contains
    all the intermediate activation values."""

    # retrieve the full execution context
    execution_context = execute_onnx(model, input_dict, True)
    new_model = copy.deepcopy(model)
    # create value_info entries and initializers for everything
    for i in execution_context.keys():
        if i != "" and execution_context[i] is not None:
            new_model.set_initializer(i, execution_context[i])
    for vi in new_model.graph.value_info:
        new_model.graph.output.append(vi)
    return new_model


def compare_execution(
    model_a,
    model_b,
    input_dict,
    compare_fxn=lambda x, y: np.isclose(x, y, atol=1e-3).all(),
):
    """Executes two ONNX models and compare their outputs using given function.

    compare_fxn should take in two tensors and return a Boolean"""
    # compare values from first output tensors produced
    res_a = list(execute_onnx(model_a, input_dict).items())[0][1]
    res_b = list(execute_onnx(model_b, input_dict).items())[0][1]
    return compare_fxn(res_a, res_b)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from onnx import TensorProto, helper
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.util.basic import calculate_signed_dot_prod_range, gen_finn_dt_tensor, qonnx_make_model


def hls_random_mlp_maker(layer_spec):
    """Create an MLP of given specification using HLSCustomOp instances.
    Generate random weights/thresholds of appropriate size."""
    ret = []
    for lyr in layer_spec:
        idt = lyr["idt"]
        wdt = lyr["wdt"]
        mw = lyr["mw"]
        mh = lyr["mh"]
        act = lyr["act"]
        lyr["W"] = gen_finn_dt_tensor(wdt, (mw, mh))
        if act is None:
            # no activation, produce accumulators
            T = None
            tdt = None
            if wdt == DataType["BIPOLAR"] and idt == DataType["BIPOLAR"]:
                odt = DataType["UINT32"]
            else:
                odt = DataType["INT32"]
        else:
            odt = act
            (min, max) = calculate_signed_dot_prod_range(idt, wdt, mw)
            n_steps = act.get_num_possible_values() - 1
            T = np.random.randint(min, max - 1, (mh, n_steps)).astype(np.float32)
            # provide non-decreasing thresholds
            T = np.sort(T, axis=1)
            # generate thresholds for activation
            if wdt == DataType["BIPOLAR"] and idt == DataType["BIPOLAR"]:
                tdt = DataType["UINT32"]
                # bias thresholds to be positive
                T = np.ceil((T + mw) / 2)
                assert (T >= 0).all()
            else:
                tdt = DataType["INT32"]
        lyr["T"] = T
        lyr["tdt"] = tdt
        lyr["odt"] = odt
        ret.append(lyr)

    return hls_mlp_maker(ret)


def hls_mlp_maker(layer_spec):
    """Create an MLP of given specification using HLSCustomOp instances."""

    current_in_name = ""
    current_out_name = ""
    i = 0

    graph = helper.make_graph(nodes=[], name="mlp", inputs=[], outputs=[])

    model = qonnx_make_model(graph, producer_name="finn")
    model = ModelWrapper(model)

    for lyr in layer_spec:
        current_W_name = "W_%d" % i
        current_T_name = "T_%d" % i
        current_in_name = "act_%d" % i
        current_out_name = "act_%d" % (i + 1)

        W = lyr["W"]
        (mw, mh) = W.shape
        T = lyr["T"]
        pe = lyr["pe"]
        simd = lyr["simd"]
        wdt = lyr["wdt"]
        idt = lyr["idt"]
        tdt = lyr["tdt"]
        odt = lyr["odt"]

        if i == 0:
            global_in = helper.make_tensor_value_info(current_in_name, TensorProto.FLOAT, [1, mw])
            model.graph.input.append(global_in)

        if i == len(layer_spec) - 1:
            global_out = helper.make_tensor_value_info(current_out_name, TensorProto.FLOAT, [1, mh])
            model.graph.output.append(global_out)

        # there are two ways to implement bipolar weights and inputs for
        # MatrixVectorActivation:
        # - specify their datatypes as such
        # - specify their datatypes as BINARY as use binaryXnorMode
        if wdt == DataType["BIPOLAR"] and idt == DataType["BIPOLAR"]:
            # we'll internally convert weights/inputs to binary and specify the
            # datatypes as such, and also set the binaryXnorMode attribute to 1
            export_wdt = DataType["BINARY"]
            export_idt = DataType["BINARY"]
            binary_xnor_mode = 1
        else:
            export_wdt = wdt
            export_idt = idt
            binary_xnor_mode = 0

        if T is not None:
            no_act = 0
            node_inp_list = [current_in_name, current_W_name, current_T_name]
            if odt == DataType["BIPOLAR"]:
                actval = 0
            else:
                actval = odt.min()
        else:
            # no thresholds
            node_inp_list = [current_in_name, current_W_name]
            actval = 0
            no_act = 1
        FCLayer_node = helper.make_node(
            "MVAU",
            node_inp_list,
            [current_out_name],
            domain="finn.custom_op.fpgadataflow",
            backend="fpgadataflow",
            MW=mw,
            MH=mh,
            SIMD=simd,
            PE=pe,
            inputDataType=export_idt.name,
            weightDataType=export_wdt.name,
            outputDataType=odt.name,
            ActVal=actval,
            binaryXnorMode=binary_xnor_mode,
            noActivation=no_act,
        )

        model.graph.node.append(FCLayer_node)
        model.set_tensor_datatype(current_in_name, idt)
        model.set_tensor_datatype(current_out_name, odt)
        model.set_tensor_datatype(current_W_name, wdt)
        if binary_xnor_mode:
            # convert bipolar to binary
            model.set_initializer(current_W_name, (W + 1) / 2)
        else:
            model.set_initializer(current_W_name, W)
        if T is not None:
            model.set_tensor_datatype(current_T_name, tdt)
            model.set_initializer(current_T_name, T)
        i += 1

    return model



# Copyright (C) 2020, Xilinx, Inc.
# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Automatically sets folding, i.e., parallelism attributes for all FINN operators."""

import functools

# Inspect information on Python objects like modules
import inspect
import numpy as np
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.transformation.general import GiveUniqueNodeNames

# Import the elementwise binary operation module to extract names of all
# specializations (which require PE parallelism to be configured)
import finn.custom_op.fpgadataflow.hls.elementwise_binary_hls as elementwise_binary_hls
from finn.analysis.fpgadataflow.dataflow_performance import dataflow_performance
from finn.transformation.fpgadataflow.annotate_cycles import AnnotateCycles
from finn.util.fpgadataflow import is_hls_node, is_rtl_node
from finn.util.logging import log


def divisors(num):
    """Yield divisors of num."""
    for x in range(1, num + 1):
        if (num % x) == 0:
            yield x


def common_divisors(numbers):
    """Return common divisors of the list of numbers."""
    separate_divisors = []
    for num in numbers:
        individual_divisors = list(divisors(num))
        separate_divisors.append(individual_divisors)

    return functools.reduce(np.intersect1d, separate_divisors)


# Find the op-type names for all HLS specializations of elementwise binary
# operations
ELEMENTWISE_BINARY_OPS = [
    op_type
    for op_type, cls in inspect.getmembers(elementwise_binary_hls, inspect.isclass)
    if issubclass(cls, elementwise_binary_hls.ElementwiseBinaryOperation_hls)
]


class SetFolding(Transformation):
    """Attempt to set parallelism attributes in all nodes to meet a specific
    target expressed as cycles per frame target_cycles_per_frame. For each
    HLSCustomOp node type, the attribute may vary but is typically one of {PE, SIMD},
    and has a certain allowed-maximum value and divisibility constraints,
    which SetFolding will take into account. Note that the algorithm implemented
    by SetFolding is very simple and it is often possible to hand-tune the returned
    parallelism configuration for better results.

    In the returned model, each node's
    cycles_estimate attribute will be set to its estimated number of cycles.

    If two_pass_relaxation is enabled,
    SetFolding will internally run a second time if the target cycles from the
    first pass could not be achieved, instead using the achievable target (which
    may be constrained by a single node) to obtain a balanced pipeline.

    Notable exceptions and special behavior:

    When folding dense convolution/FC compute engines ("MVAU"/MatrixVectorActivation),
    which have two attributes (PE and SIMD):

    * first increases SIMD while weight stream width per PE is <= mvau_wwidth_max
      (configurable in the SetFolding initializer, defaults to 36)
    * then increases PE until the target is met or max PE reached

    When folding depthwise convolutions ("VVAU"/VectorVectorActivation)
    or spatial reduction ops (Pool_Batch):

    * the producer of the node is expected to be a ConvolutionInputGenerator
      with depthwise=1, whose SIMD value will be set equal to the PE value of
      its consumer node
    * the VVAU also supports SIMD ("input window") parallelism next to
      PE ("channels"), but current ConvInpGen limitations require PE to be fully
      unfolded before SIMD is increased
    """

    def __init__(self, target_cycles_per_frame=1000, mvau_wwidth_max=36, two_pass_relaxation=True):
        """Initializes the folding target and constraints."""
        super().__init__()
        self.target_cycles_per_frame = target_cycles_per_frame
        self.mvau_wwidth_max = mvau_wwidth_max
        self.two_pass_relaxation = two_pass_relaxation

    def optimize_attribute_val(self, node_inst, max_val, attr_name):
        """Optimize the folding attribute until the target cycles are met."""
        node_inst.set_nodeattr(attr_name, 1)
        for val in divisors(max_val):
            node_inst.set_nodeattr(attr_name, val)
            cyc = node_inst.get_exp_cycles()
            if cyc <= self.target_cycles_per_frame:
                # finish if target met
                break

    def apply(self, model):
        """Apply SetFolding to all supported nodes in the model."""
        graph = model.graph
        # these ops use PE parallelism, up to a max value of NumChannels
        pe_ops = [
            "AddStreams_hls",
            "ChannelwiseOp_hls",
            "DuplicateStreams_hls",
            "GlobalAccPool_hls",
            "Thresholding_hls",
            "Thresholding_rtl",
            "ReplicateStream_hls",
            *ELEMENTWISE_BINARY_OPS,
            "Squeeze_hls",
            "Unsqueeze_hls",
            "Reshape_rtl",
        ]
        # these ops use SIMD parallelism, up to a max value of NumChannels
        # ConvolutionInputGenerator has a special case when depthwise=1
        # ConvolutionInputGenerator_rtl supports additional parallelism by
        # setting parallel_window=1 mode after maxing out SIMD
        simd_ops = [
            "FMPadding_rtl",
            "FMPadding_Pixel_hls",
            "ConvolutionInputGenerator_rtl",
            # Streaming Split and Concat are SIMD operations
            "StreamingSplit_hls",
            "StreamingConcat_hls",
        ]
        # these ops are preceded by depthwise SWG and have special behavior,
        # as explained in the SetFolding docstring
        depthwise_op_exceptions = ["VVAU_hls", "VVAU_rtl", "Pool_hls"]
        for node in graph.node:
            if not (is_hls_node(node) or is_rtl_node(node)):
                continue
            op_type = node.op_type
            node_inst = getCustomOp(node)
            if op_type in ["MVAU_hls", "MVAU_rtl"]:
                max_simd = node_inst.get_nodeattr("MW")
                max_pe = node_inst.get_nodeattr("MH")
                node_inst.set_nodeattr("PE", 1)
                node_inst.set_nodeattr("SIMD", 1)
                # increase SIMD until either we meet
                # the target or weight stream becomes
                # too wide
                for simd_val in divisors(max_simd):
                    prev_simd_val = node_inst.get_nodeattr("SIMD")
                    node_inst.set_nodeattr("SIMD", simd_val)
                    cyc = node_inst.get_exp_cycles()
                    if cyc <= self.target_cycles_per_frame:
                        # finish if target met
                        break
                    if (
                        node_inst.get_input_datatype(1).bitwidth() * node_inst.get_nodeattr("SIMD")
                        > self.mvau_wwidth_max
                    ):
                        # revert if we've gone above width threshold
                        node_inst.set_nodeattr("SIMD", prev_simd_val)
                        break
                # increase PE until target met or reached max_pe
                self.optimize_attribute_val(node_inst, max_pe, "PE")
            elif op_type in pe_ops:
                # Note: Keep original behavior for all custom-ops defining the
                # NumChannels attribute as it is
                try:
                    max_pe = node_inst.get_nodeattr("NumChannels")
                # Note: Some of the recent additions do not define the
                # NumChannels attribute
                except AttributeError:
                    # We can extract the channels from the normal, i.e., not
                    # folded, shape of the output in these cases
                    max_pe = node_inst.get_normal_output_shape()[-1]
                self.optimize_attribute_val(node_inst, max_pe, "PE")
            elif op_type == "LabelSelect_hls":
                max_pe = node_inst.get_nodeattr("Labels")
                self.optimize_attribute_val(node_inst, max_pe, "PE")
            elif op_type in depthwise_op_exceptions:
                # init/reset SIMD of VVAU
                if op_type in ["VVAU_hls", "VVAU_rtl"]:
                    node_inst.set_nodeattr("SIMD", 1)
                max_pe = node_inst.get_nodeattr("Channels")
                self.optimize_attribute_val(node_inst, max_pe, "PE")
                # increase SIMD for VVAU once PE is exhausted
                pe = node_inst.get_nodeattr("PE")
                cyc = node_inst.get_exp_cycles()
                if (
                    op_type in ["VVAU_hls", "VVAU_rtl"]
                    and pe == max_pe
                    and cyc > self.target_cycles_per_frame
                ):
                    max_simd = np.prod(node_inst.get_nodeattr("Kernel"))
                    self.optimize_attribute_val(node_inst, max_simd, "SIMD")
                # also set the folding of the upsteam DW SWU
                # which must be identical to this node
                swu_node = model.find_producer(node.input[0])
                if swu_node.op_type.startswith("ConvolutionInputGenerator"):
                    swu_node_inst = getCustomOp(swu_node)
                    swu_node_inst.set_nodeattr("SIMD", pe)
                    # enable parallel_window mode of RTL SWG if needed
                    if swu_node.op_type == "ConvolutionInputGenerator_rtl":
                        if op_type.startswith("VVAU") and node_inst.get_nodeattr("SIMD") > 1:
                            swu_node_inst.set_nodeattr("parallel_window", 1)
                        else:
                            swu_node_inst.set_nodeattr("parallel_window", 0)
                else:
                    if op_type in ["VVAU_hls", "VVAU_rtl"]:
                        ksize = np.prod(node_inst.get_nodeattr("Kernel"))
                    elif op_type == "Pool_hls":
                        ksize = node_inst.get_nodeattr("KernelSize")
                    else:
                        raise Exception("Undefined edge case for %s" % op_type)
                    if ksize != 1:  # pointwise vvau/pool lack a SWU
                        raise Exception("Expected SWU on DW op input, found " + swu_node.op_type)
            elif op_type in simd_ops:
                if op_type.startswith("ConvolutionInputGenerator"):
                    depthwise = node_inst.get_nodeattr("depthwise")
                    if depthwise == 0:
                        max_simd = node_inst.get_nodeattr("IFMChannels")
                        # init/reset parallel_window mode of RTL SWG
                        if op_type == "ConvolutionInputGenerator_rtl":
                            node_inst.set_nodeattr("parallel_window", 0)
                        self.optimize_attribute_val(node_inst, max_simd, "SIMD")
                        # enable parallel_window mode of RTL SWG if needed
                        simd = node_inst.get_nodeattr("SIMD")
                        cyc = node_inst.get_exp_cycles()
                        if (
                            op_type == "ConvolutionInputGenerator_rtl"
                            and simd == max_simd
                            and cyc > self.target_cycles_per_frame
                        ):
                            node_inst.set_nodeattr("parallel_window", 1)
                    else:
                        # depthwise SWGs are handled separately
                        continue
                elif op_type == "StreamingConcat_hls" or op_type == "StreamingSplit_hls":
                    node_inst.set_nodeattr("SIMD", 1)
                    channels_per_stream = node_inst.get_nodeattr("ChannelsPerStream")
                    for simd_val in common_divisors(channels_per_stream):
                        node_inst.set_nodeattr("SIMD", simd_val)
                        cyc = node_inst.get_exp_cycles()
                        if cyc < self.target_cycles_per_frame:
                            break
                else:
                    # Note: Keep original behavior for all custom-ops defining
                    # the NumChannels attribute as it is
                    try:
                        max_simd = node_inst.get_nodeattr("NumChannels")
                    # Note: Some of the recent additions do not define the
                    # NumChannels attribute
                    except AttributeError:
                        # We can extract the channels from the normal, i.e., not
                        # folded, shape of the input in these cases
                        max_simd = node_inst.get_normal_input_shape()[-1]
                    self.optimize_attribute_val(node_inst, max_simd, "SIMD")
            else:
                log.warning(f"SetFolding doesn't know how to handle op_type {op_type}")

        model = model.transform(GiveUniqueNodeNames())
        model = model.transform(AnnotateCycles())
        if self.two_pass_relaxation:
            perf_dict = model.analysis(dataflow_performance)
            if perf_dict["max_cycles"] > self.target_cycles_per_frame:
                # run again, but with lower target (that we managed) -- this
                # may be coming from a single node's constraints, but we want
                # to balance the entire dataflow pipeline instead
                # no two_pass_relaxation this time -- no guarantee we'll
                # converge otherwise
                max_cycles_node_name = perf_dict["max_cycles_node_name"]
                max_cycles = perf_dict["max_cycles"]
                log.warning(
                    f"Node {max_cycles_node_name} is bottleneck with {max_cycles} cycles, \
                        running second pass"
                )
                model = model.transform(
                    SetFolding(
                        target_cycles_per_frame=perf_dict["max_cycles"],
                        mvau_wwidth_max=self.mvau_wwidth_max,
                        two_pass_relaxation=False,
                    )
                )

        return (model, False)



# Copyright (c) 2021, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from qonnx.transformation.base import Transformation
from qonnx.util.basic import get_by_name


class ExternalizeParams(Transformation):
    """Create top-level graph inputs for IODMAs serving layers where weights are
    marked as external using mem_mode="external"."""

    def __init__(self):
        super().__init__()

    def apply(self, model):
        graph_modified = False

        def filter_fc_extw(x):
            if x.op_type == "IODMA_hls":
                burst_mode = get_by_name(x.attribute, "burstMode")
                if burst_mode is not None:
                    burst_mode = burst_mode.s.decode("UTF-8")
                    return burst_mode == "wrap"

        dma_extw_nodes = list(filter(filter_fc_extw, model.graph.node))

        for dma_extw in dma_extw_nodes:
            extw_tensor_name = dma_extw.input[0]
            extw_tensor_name_out = dma_extw.output[0]
            if extw_tensor_name in [x.name for x in model.graph.input]:
                continue
            else:
                extw_vi = model.get_tensor_valueinfo(extw_tensor_name)
                assert extw_vi is not None
                model.graph.value_info.remove(extw_vi)
                model.graph.input.append(extw_vi)
                iodma_init = model.get_initializer(extw_vi.name)
                assert iodma_init is not None
                # remove output-side initializer to get correct dataflow partitioning
                model.graph.initializer.remove(
                    [x for x in model.graph.initializer if x.name == extw_tensor_name_out][0]
                )
                graph_modified = True

        return (model, graph_modified)



# Copyright (c) 2021, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from qonnx.transformation.base import Transformation

from finn.transformation.qonnx.qonnx_activation_handlers import (
    QuantActBaseHandler,
    QuantIdentityHandler,
)
from finn.util.logging import log


def default_filter_function_generator(max_multithreshold_bit_width=8):
    """
    This function generates the default filter function for the
    ConvertQuantActToMultiThreshold transformation. Per default the returned
    function disables the conversion of Quant nodes which have a bit width above 8 bit.

    This function generator can be used as a template to write custom
    filter functions.
    """

    def filter_function(model, q_node):
        if q_node.op_type == "Quant":
            bit_width = model.get_initializer(q_node.input[3])
        elif q_node.op_type == "BipolarQuant":
            bit_width = 1.0
        else:
            raise RuntimeError("Got an unexpected quantizer node type")
        if bit_width is None:
            raise ValueError("Quant nodes must have a static bit width.")
        if bit_width > max_multithreshold_bit_width:
            log.warning(
                f'The Quant node with name: "{q_node.name}" was not converted to a '
                f"MultiThreshold node, because its bit width of {bit_width} is "
                f"higher than the configured maximum bit width of "
                f"{max_multithreshold_bit_width}."
            )
            return False
        return True

    return filter_function


class ConvertQuantActToMultiThreshold(Transformation):
    """Converts Quant nodes in the activation path to MultiThreshold nodes.

    The optional keyword argument `filter_function`
    presents a way to control which Quant and BipolarQuant nodes in the activation path
    are converted to MultiThreshold nodes. A warning will be emitted when a Quant node
    is not converted to a MultiThreshold node.

    :param filter_function: Each candidate Quant and BinaryQant node is first evaluated
        by this function. If the function returns False,
        then the node is not converted to a MultiTrheshold node.
        The function is given the model and candidate node as parameters.
        Per default a filter function is inserted, which disables the conversion of
        Quant nodes, which have a bit width of larger than 8.
        Defaults to: default_filter_function_generator(max_multithreshold_bit_width=8)
    """

    def __init__(
        self,
        filter_function=default_filter_function_generator(max_multithreshold_bit_width=8),
    ):
        super().__init__()
        self._filter_function = filter_function

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        graph_modified = False

        for n in graph.node:
            node_ind += 1
            if n.op_type == "Quant" or n.op_type == "BipolarQuant":
                # Check that the node is in the activation path
                inp = model.get_initializer(n.input[0])
                out = model.get_initializer(n.output[0])
                if not (inp is None and out is None):
                    continue
                predecessor = model.find_direct_predecessors(n)
                if predecessor is not None:
                    predecessor_op_type = predecessor[0].op_type
                else:
                    predecessor_op_type = predecessor
                if n.op_type == "Quant" and not model.get_initializer(n.input[2]) == 0:
                    raise ValueError(
                        "Only Quant nodes with zero-point == 0 are currently supported."
                    )

                # Check that this node passes the user filter
                if not self._filter_function(model, n):
                    log.warning(
                        f'The Quant node with name: "{n.name}" was not converted to a '
                        f"MultiThreshold node, because the filtering function "
                        f"returned False for this node."
                    )
                    continue

                # Check for possible ambiguity in handler selection
                valid_predecessors = []
                for cls in QuantActBaseHandler.__subclasses__():
                    valid_predecessors.extend(cls.valid_predecessor_op_types())
                if len(valid_predecessors) != len(set(valid_predecessors)):
                    raise RuntimeError(
                        "Two or more activation handlers declare the same "
                        "type of valid predecessor node. "
                        "This leads to ambiguity in the handler selection "
                        "and must thus be avoided."
                    )

                # Try to find a fitting handler for this Quant activation node
                for handler_cls in QuantActBaseHandler.__subclasses__():
                    if predecessor_op_type in handler_cls.valid_predecessor_op_types():
                        handler = handler_cls(model, n, node_ind)
                        break
                else:
                    # fall back to QuantIdentityHandler here
                    # it may still not work due to its particular restrictions,
                    # but better than just erroring out without trying
                    handler = QuantIdentityHandler(model, n, node_ind)

                model = handler.replace_quant_node()
                graph_modified = True
                return (model, graph_modified)

        return (model, graph_modified)



"""Analysis for detecting unsupported layers in FINN dataflow models.

This module provides functionality to validate that FINN models have a valid
partition structure where FPGA-supported operations form a single contiguous
section in the dataflow graph.
"""

from collections import deque
from qonnx.core.modelwrapper import ModelWrapper


def unsupported_layers(model: ModelWrapper):
    """
    Check if all sink nodes are only reachable by paths with at most one
    connected section of nodes which are supported by the FPGA.
    """

    def is_supported_node(node):
        """Check if a node is supported by (= mapped to) the FPGA backend."""
        return node.domain.startswith("finn.custom_op.fpgadataflow")

    # Find source and sink nodes in the model
    source_nodes = []
    sink_nodes = []

    inputs = model.graph.input
    for inp in inputs:
        # Nodes are not supported by python sets, so needing to do deduplication manually
        for n in model.find_consumers(inp.name):
            if n not in source_nodes:
                source_nodes.append(n)

    outputs = model.graph.output
    for out in outputs:
        n = model.find_producer(out.name)
        if n not in sink_nodes:
            sink_nodes.append(n)

    # BFS to check paths
    queue = deque()
    # Track (node_id, in_green_section, has_seen_complete_green_section)
    visited = []

    # Initialize BFS with all source nodes
    for source in source_nodes:
        is_supported = is_supported_node(source)
        queue.append((source, is_supported, False))
        visited.append((source, is_supported, False))

    while queue:
        node, in_fpga_sec, seen_complete_fpga_section = queue.popleft()

        # Process all successors
        successors = model.find_direct_successors(node)
        if successors is not None:
            for successor in successors:
                sucessor_supported = is_supported_node(successor)
                fpga_section_end = seen_complete_fpga_section

                # If transitioning from FPGA to Host, we've completed a FPGA section
                if in_fpga_sec and not sucessor_supported:
                    fpga_section_end = True

                # If we've already seen a complete FPGA section and are about to start a new one
                if seen_complete_fpga_section and not in_fpga_sec and sucessor_supported:
                    # This path would create a second FPGA section
                    return False, node

                # Handle cycles
                state = (successor, sucessor_supported, fpga_section_end)
                if state not in visited:
                    visited.append(state)
                    queue.append((successor, sucessor_supported, fpga_section_end))

    return True, None



#############################################################################
# Copyright (C) 2025, Advanced Micro Devices, Inc.
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause
#
# @brief	rtlsim_multi_io interface adapter for FINN XSI
# @author	Yaman Umuroglu <yaman.umuroglu@amd.com>
#############################################################################

import errno
import os
import os.path
import re
from finn_xsi.sim_engine import SimEngine
from typing import Optional

from finn.util.basic import launch_process_helper


def locate_glbl() -> Optional[str]:
    """
    Tries to determine the glbl.v file path from environment variables.
    Returns None if it cannot be found.
    """
    # Get GLBL from the Vitis environment variable
    vivado_path = os.environ.get("XILINX_VIVADO")
    if vivado_path:
        glbl_path = os.path.join(vivado_path, "data", "verilog", "src", "glbl.v")
        if os.path.isfile(glbl_path):
            return glbl_path
    return None


def compile_sim_obj(top_module_name, source_list, sim_out_dir, debug=False):
    # create a .prj file with the source files
    with open(sim_out_dir + "/rtlsim.prj", "w") as f:
        glbl = locate_glbl()
        if glbl is not None:
            f.write(f"verilog work {glbl}\n")

        # extract (unique, by using a set) verilog headers for inclusion
        verilog_headers = {os.path.dirname(x) for x in source_list if x.endswith(".vh")}
        verilog_header_incl_str = " ".join(["--include " + x for x in verilog_headers])

        for src_line in source_list:
            if src_line.endswith(".v"):
                f.write(f"verilog work {verilog_header_incl_str} {src_line}\n")
            elif src_line.endswith(".vhd"):
                # note that Verilog header incls are not added for VHDL
                f.write(f"vhdl2008 work {src_line}\n")
            elif src_line.endswith(".sv"):
                f.write(f"sv work {verilog_header_incl_str} {src_line}\n")
            elif src_line.endswith(".vh"):
                # skip adding Verilog headers directly (see verilog_header_incl_str)
                continue
            else:
                raise Exception(f"Unknown extension for .prj file sources: {src_line}")

    # now call xelab to generate the .so for the design to be simulated
    # list of libs for xelab retrieved from Vitis HLS cosim cmdline
    # the particular lib version used depends on the Vivado/Vitis version being used
    # but putting in multiple (nonpresent) versions seems to pose no problem as long
    # as the correct one is also in there. at least this is how Vitis HLS cosim is
    # handling it.
    # TODO make this an optional param instead of hardcoding
    xelab_libs = [
        "smartconnect_v1_0",
        "axi_protocol_checker_v1_1_12",
        "axi_protocol_checker_v1_1_13",
        "axis_protocol_checker_v1_1_11",
        "axis_protocol_checker_v1_1_12",
        "xil_defaultlib",
        "unisims_ver",
        "xpm",
        "floating_point_v7_1_16",
        "floating_point_v7_0_21",
        "floating_point_v7_1_18",
        "floating_point_v7_1_15",
        "floating_point_v7_1_19",
    ]

    cmd_xelab = [
        "xelab",
        "work." + top_module_name,
        "-relax",
        "-prj",
        "rtlsim.prj",
        "-dll",
        "-s",
        top_module_name,
    ]
    # Add debug flag if debug is enabled
    if debug:
        cmd_xelab.append("-debug")
        cmd_xelab.append("all")
    for lib in xelab_libs:
        cmd_xelab.append("-L")
        cmd_xelab.append(lib)

    if locate_glbl() is not None:
        cmd_xelab.insert(1, "work.glbl")

    launch_process_helper(cmd_xelab, cwd=sim_out_dir)
    out_so_relative_path = "xsim.dir/%s/xsimk.so" % top_module_name
    out_so_full_path = sim_out_dir + "/" + out_so_relative_path

    if not os.path.isfile(out_so_full_path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), out_so_full_path)

    return (sim_out_dir, out_so_relative_path)


def get_simkernel_so():
    vivado_path = os.environ.get("XILINX_VIVADO")
    # xsi kernel lib name depends on Vivado version (renamed in 2024.2)
    match = re.search(r"\b(20\d{2})\.(1|2)\b", vivado_path)
    year, minor = int(match.group(1)), int(match.group(2))
    if (year, minor) > (2024, 1):
        simkernel_so = "libxv_simulator_kernel.so"
    else:
        simkernel_so = "librdi_simulator_kernel.so"
    return simkernel_so


def load_sim_obj(sim_out_dir, out_so_relative_path, tracefile=None, simkernel_so=None):
    if simkernel_so is None:
        simkernel_so = get_simkernel_so()
    oldcwd = os.getcwd()
    os.chdir(sim_out_dir)
    sim = SimEngine(simkernel_so, out_so_relative_path, "finnxsi_rtlsim.log", tracefile)
    if tracefile:
        sim.top.trace_all()
    os.chdir(oldcwd)
    return sim


def reset_rtlsim(
    sim, rst_name="ap_rst_n", active_low=True, clk_name="ap_clk", clk2x_name="ap_clk2x", n_cycles=16
):
    sim.do_reset()
    sim.run()


def close_rtlsim(sim):
    del sim


def rtlsim_multi_io(
    sim,
    io_dict,
    num_out_values,
    sname="_V_V",
    liveness_threshold=10000,
):
    if len(io_dict["outputs"]) > 1:
        assert isinstance(
            num_out_values, dict
        ), "num_out_values must be dict for multiple output streams"
    else:
        # num_out_values is provided as integer (indicating the expected
        # outputs from the single output stream) - make into dict
        oname = list(io_dict["outputs"].keys())[0]
        num_out_values = {oname: num_out_values}
    # FINN XSI expects hex strings, while rtlsim_multi_io uses
    # lists of arbitrary-precision integers, so need to convert
    # inputs and outputs to appropriate format
    # TODO: refactor components&data packing to directly generate and consume
    # hex strings instead of arb-prec Python integers
    for inp in io_dict["inputs"]:
        arbprec_int_input = io_dict["inputs"][inp]
        hexstring_input = map(lambda var: f"{var:0x}", arbprec_int_input)
        stream_name = inp + sname
        sim.stream_input(stream_name, hexstring_input)

    hex_output_streams = {}
    for out in io_dict["outputs"]:
        stream_name = out + sname
        hex_output_streams[out] = sim.collect_output(
            stream_name,
            num_out_values[out],
            watchdog=sim.create_watchdog(f"{stream_name} timeout", liveness_threshold),
        )

    start_ticks = sim.ticks
    ret = sim.run()
    if len(ret) > 0:
        assert False, f"RTL simulation watchdogs {str(ret)} timed out. Check rtlsim_trace if any."
    end_ticks = sim.ticks
    for out in io_dict["outputs"]:
        io_dict["outputs"][out] = list(map(lambda var: int(var, base=16), hex_output_streams[out]))

    return end_ticks - start_ticks



# Copyright (c) 2020, Xilinx
# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from qonnx.custom_op.registry import getCustomOp

from finn.util.fpgadataflow import is_fpgadataflow_node


def floorplan_params(model):
    """Gathers SLR and partition IDs from nodes.

    Returns {node name : {slr, device id, partition id, memory port}}."""

    ret_dict = {
        "Defaults": {
            "slr": [-1, ["all"]],
            "partition_id": [0, ["all"]],
            "device_id": [0, ["all"]],
            "mem_port": ["", ["all"]],
        }
    }
    for node in model.graph.node:
        if is_fpgadataflow_node(node):
            node_inst = getCustomOp(node)
            node_slr = node_inst.get_nodeattr("slr")
            node_pid = node_inst.get_nodeattr("partition_id")
            node_mport = node_inst.get_nodeattr("mem_port")
            ret_dict[node.name] = {
                "slr": node_slr,
                "partition_id": node_pid,
                "device_id": 0,
                "mem_port": node_mport,
            }

    return ret_dict



# Copyright (C) 2020-2022, Xilinx, Inc.
# Copyright (C) 2023-2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp

# Dictionary of HWCustomOp implementations
custom_op = dict()


# Registers a class into the custom_op dictionary
# Note: This must be defined first, before importing any custom op
# implementation to avoid "importing partially initialized module" issues.
def register_custom_op(cls):
    # The class must actually implement HWCustomOp
    assert issubclass(cls, HWCustomOp), f"{cls} must subclass {HWCustomOp}"
    # Insert the class into the custom_op dictionary by its name
    custom_op[cls.__name__] = cls
    # Pass through the class unmodified
    return cls


# flake8: noqa
# Disable linting from here, as all import will be flagged E402 and maybe F401


# Import the submodule containing specializations of ElementwiseBinaryOperation
# Note: This will automatically register all decorated classes into this domain
import finn.custom_op.fpgadataflow.elementwise_binary

# Import the submodule containing the Squeeze operation
# Note: This will automatically register all decorated classes into this domain
import finn.custom_op.fpgadataflow.squeeze

# Import the submodule containing the Unsqueeze operation
import finn.custom_op.fpgadataflow.unsqueeze
from finn.custom_op.fpgadataflow.addstreams import AddStreams
from finn.custom_op.fpgadataflow.attention import ScaledDotProductAttention
from finn.custom_op.fpgadataflow.attention_heads import MergeMultiHeads, SplitMultiHeads
from finn.custom_op.fpgadataflow.channelwise_op import ChannelwiseOp
from finn.custom_op.fpgadataflow.concat import StreamingConcat
from finn.custom_op.fpgadataflow.convolutioninputgenerator import ConvolutionInputGenerator
from finn.custom_op.fpgadataflow.duplicatestreams import DuplicateStreams
from finn.custom_op.fpgadataflow.fmpadding import FMPadding
from finn.custom_op.fpgadataflow.fmpadding_pixel import FMPadding_Pixel
from finn.custom_op.fpgadataflow.globalaccpool import GlobalAccPool
from finn.custom_op.fpgadataflow.labelselect import LabelSelect
from finn.custom_op.fpgadataflow.lookup import Lookup
from finn.custom_op.fpgadataflow.matrixvectoractivation import MVAU
from finn.custom_op.fpgadataflow.pool import Pool
from finn.custom_op.fpgadataflow.replicate_stream import ReplicateStream
from finn.custom_op.fpgadataflow.split import StreamingSplit
from finn.custom_op.fpgadataflow.streamingdataflowpartition import StreamingDataflowPartition
from finn.custom_op.fpgadataflow.streamingdatawidthconverter import StreamingDataWidthConverter
from finn.custom_op.fpgadataflow.streamingeltwise import StreamingEltwise
from finn.custom_op.fpgadataflow.streamingfifo import StreamingFIFO
from finn.custom_op.fpgadataflow.thresholding import Thresholding
from finn.custom_op.fpgadataflow.upsampler import UpsampleNearestNeighbour
from finn.custom_op.fpgadataflow.vectorvectoractivation import VVAU

# make sure new HLSCustomOp subclasses are imported here so that they get
# registered and plug in correctly into the infrastructure
custom_op["MVAU"] = MVAU
custom_op["StreamingFIFO"] = StreamingFIFO
custom_op["Thresholding"] = Thresholding
custom_op["VVAU"] = VVAU
custom_op["StreamingDataflowPartition"] = StreamingDataflowPartition

custom_op["AddStreams"] = AddStreams
custom_op["ChannelwiseOp"] = ChannelwiseOp
custom_op["ConvolutionInputGenerator"] = ConvolutionInputGenerator
custom_op["DuplicateStreams"] = DuplicateStreams
custom_op["FMPadding"] = FMPadding
custom_op["FMPadding_Pixel"] = FMPadding_Pixel
custom_op["GlobalAccPool"] = GlobalAccPool
custom_op["LabelSelect"] = LabelSelect
custom_op["Lookup"] = Lookup
custom_op["Pool"] = Pool
custom_op["StreamingConcat"] = StreamingConcat
custom_op["StreamingSplit"] = StreamingSplit
custom_op["StreamingDataWidthConverter"] = StreamingDataWidthConverter
custom_op["StreamingEltwise"] = StreamingEltwise
custom_op["UpsampleNearestNeighbour"] = UpsampleNearestNeighbour
custom_op["ScaledDotProductAttention"] = ScaledDotProductAttention
custom_op["SplitMultiHeads"] = SplitMultiHeads
custom_op["MergeMultiHeads"] = MergeMultiHeads
custom_op["ReplicateStream"] = ReplicateStream



# Copyright (c) 2022, Xilinx
# Copyright (C) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from qonnx.core.datatype import DataType

from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp
from finn.util.logging import log


class StreamingEltwise(HWCustomOp):
    """Abstraction layer for HW implementation of StreamingEltwise"""

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = super().get_nodeattr_types()
        my_attrs.update(
            {
                "NumChannels": ("i", True, ""),
                "PE": ("i", True, ""),
                # FINN DataTypes for inputs; output datatype inferred from input
                "inputDataType0": ("s", True, ""),
                "inputDataType1": ("s", True, ""),
                # type of EltwiseFunction for the operation
                "eltwiseOp": ("s", True, "", ["Add", "Sub", "AbsDiff"]),
                # number of input vectors, examples:
                # [1] is a single vector (like a FC layer with batch=1)
                # [4] is four vectors (like a FC layer with batch=4)
                # [1, 4, 4] is four * four vectors (like a conv layer with batch=1)
                "numInputVectors": ("ints", False, [1]),
                "inFIFODepths": ("ints", False, [2, 2]),
            }
        )
        return my_attrs

    def get_eltwise_op_lambda(self):
        eltwise_op = self.get_nodeattr("eltwiseOp")
        idt0 = self.get_input_datatype(0)
        idt1 = self.get_input_datatype(1)
        odt = self.get_output_datatype()
        tin0 = idt0.get_hls_datatype_str()
        tin1 = idt1.get_hls_datatype_str()
        tout = odt.get_hls_datatype_str()
        eltwise_ops = {
            # "Add": "[](auto a, auto b) { return  a + b; }",
            # "Sub": "[](auto a, auto b) { return  a - b; }",
            # "AbsDiff": "[](auto a, auto b) { return  a>b? a-b : b-a; }",
            "Add": f"add<{tin0}, {tin1}, {tout}>()",
            "Sub": f"sub<{tin0}, {tin1}, {tout}>()",
            "AbsDiff": f"absdiff<{tin0}, {tin1}, {tout}>()",
        }
        return eltwise_ops[eltwise_op]

    def get_normal_input_shape(self, ind=0):
        ich = self.get_nodeattr("NumChannels")
        vecs = list(self.get_nodeattr("numInputVectors"))
        ishape = tuple(vecs + [ich])
        return ishape

    def get_folded_input_shape(self, ind=0):
        ich = self.get_nodeattr("NumChannels")
        pe = self.get_nodeattr("PE")
        assert ich % pe == 0, "PE must divide NumChannels"
        vecs = list(self.get_nodeattr("numInputVectors"))
        ishape = tuple(vecs + [ich // pe, pe])
        return ishape

    def get_normal_output_shape(self, ind=0):
        return self.get_normal_input_shape()

    def get_folded_output_shape(self, ind=0):
        return self.get_folded_input_shape()

    def infer_node_datatype(self, model):
        node = self.onnx_node
        idt0 = model.get_tensor_datatype(node.input[0])
        if idt0 != self.get_input_datatype(0):
            warn_str = "inputDataType0 changing for %s: %s -> %s " % (
                node.name,
                str(self.get_input_datatype(0)),
                str(idt0),
            )
            log.warning(warn_str)
        self.set_nodeattr("inputDataType0", idt0.name)
        idt1 = model.get_tensor_datatype(node.input[1])
        if idt1 != self.get_input_datatype(1):
            warn_str = "inputDataType1 changing for %s: %s -> %s " % (
                node.name,
                str(self.get_input_datatype(1)),
                str(idt1),
            )
            log.warning(warn_str)
        self.set_nodeattr("inputDataType1", idt1.name)
        # enforce output data type (calculated based on idt)
        odt = self.get_output_datatype()
        model.set_tensor_datatype(self.onnx_node.output[0], odt)

    def get_input_datatype(self, ind=0):
        """Returns FINN DataType of input."""
        return DataType[self.get_nodeattr("inputDataType" + str(ind))]

    def get_output_datatype(self, ind=0):
        """Returns FINN DataType of output."""
        op = self.get_nodeattr("eltwiseOp")
        idt0 = self.get_input_datatype(0)
        idt1 = self.get_input_datatype(1)
        assert idt0.signed() == idt1.signed(), (
            "%s: Inputs must have same signedness" % self.onnx_node.name
        )
        idt0_min, idt0_max = idt0.min(), idt0.max()
        idt1_min, idt1_max = idt1.min(), idt1.max()
        cands = [
            idt0_min - idt1_min,
            idt0_min - idt1_max,
            idt0_max - idt1_min,
            idt0_max - idt1_max,
        ]
        largest_magnitude = max(map(abs, cands))
        if op == "Add":
            if idt0.signed():
                return DataType.get_smallest_possible(idt0.min() + idt1.min())
            else:
                return DataType.get_smallest_possible(idt0.max() + idt1.max())
        elif op == "Sub":
            return DataType.get_smallest_possible(-largest_magnitude)
        elif op == "AbsDiff":
            return DataType.get_smallest_possible(largest_magnitude)
        else:
            raise Exception("%s: Unknown eltWiseOp = %s" % (self.onnx_node.name, op))

    def get_instream_width(self, ind=0):
        """Returns input stream width."""
        ibits = self.get_input_datatype(ind).bitwidth()
        pe = self.get_nodeattr("PE")
        in_width = pe * ibits
        return in_width

    def get_outstream_width(self, ind=0):
        """Returns output stream width."""
        obits = self.get_output_datatype().bitwidth()
        pe = self.get_nodeattr("PE")
        out_width = pe * obits
        return out_width

    def get_exp_cycles(self):
        # Channels/PE * batch size * fmdim * fmdim
        return np.prod(self.get_folded_output_shape()[:-1])

    def execute_node(self, context, graph):
        # simulate behavior using Python
        node = self.onnx_node
        inp0_values = context[node.input[0]]
        inp1_values = context[node.input[1]]
        eltwiseOp = self.get_nodeattr("eltwiseOp")
        oshape = context[node.output[0]].shape
        ishape0 = inp0_values.shape
        ishape1 = inp1_values.shape
        assert ishape0 == ishape1, "Shapes of inputs should be the same for Streamingeltwise"
        # subtraction
        result = inp0_values - inp1_values
        if eltwiseOp == "Sub":
            context[node.output[0]] = np.asarray(result, dtype=np.float32).reshape(oshape)
        elif eltwiseOp == "AbsDiff":
            context[node.output[0]] = np.abs(np.asarray(result, dtype=np.float32)).reshape(oshape)
        else:
            raise Exception("%s: Unknown eltWiseOp = %s" % (node.name, eltwiseOp))

    def get_verilog_top_module_intf_names(self):
        intf_names = super().get_verilog_top_module_intf_names()
        swidth = self.get_instream_width_padded()
        intf_names["s_axis"] = [(x + "_V", swidth) for x in ["in0", "in1"]]
        return intf_names



"""Utilities for the FINN command line interface."""

from __future__ import annotations

import multiprocessing
import os
import sys
import yaml
from pathlib import Path
from rich.console import Console

from finn.interface import DEBUG


def error(msg: str) -> None:
    """Print an error"""
    Console().print(f"[bold red]ERROR: [/bold red][red]{msg}[/red]")


def warning(msg: str) -> None:
    """Print a warning"""
    Console().print(f"[bold orange1]WARNING: [/bold orange1][orange3]{msg}[/orange3]")


def status(msg: str) -> None:
    """Print a status message"""
    Console().print(f"[bold cyan]STATUS: [/bold cyan][cyan]{msg}[/cyan]")


def success(msg: str) -> None:
    """Print a success message"""
    Console().print(f"[bold green]SUCCESS: [/bold green][green]{msg}[/green]")


def debug(msg: str) -> None:
    """Print a debug message. Only done when the flag is set"""
    if DEBUG:
        Console().print(f"[bold blue]DEBUG: [/bold blue][blue]{msg}[/blue]")


def assert_path_valid(p: Path) -> None:
    """Check if the path exists, if not print an error message and exit with an error code"""
    if not p.exists():
        Console().print(f"[bold red]File or directory {p} does not exist. Stopping...[/bold red]")
        sys.exit(1)


def set_synthesis_tools_paths() -> None:
    """Check that all synthesis tools can be found. If not, give a warning."""
    for envname, toolname in [
        ("XILINX_VIVADO", "vivado"),
        ("XILINX_VITIS", "vitis"),
        ("XILINX_HLS", "vitis_hls"),
    ]:
        if envname not in os.environ.keys():
            warning(
                f"Path to the {toolname} tool could not be resolved from {envname}. "
                "Did you source your settings file?"
            )
            continue
        envname_path = os.environ[envname]

        # Exception for Vitis HLS because of changed behavior starting with 2024.2
        # XILINX_HLS no longer points to */Vitis_HLS/VERSION but */Vitis/VERSION
        p = Path(envname_path) / "bin" / toolname
        if not p.exists() and toolname == "vitis_hls":
            envname_path = envname_path.replace("Vitis", "Vitis_HLS")
            p = Path(envname_path) / "bin" / toolname

        if not p.exists():
            warning(f"Path for {toolname} found, but executable not found in {p}!")
        # TODO: simply check "which" instead?

        # Append to the search path just in case it was missing before, if
        # already in there adding again should do nothing
        os.environ["PATH"] += os.pathsep + str((Path(envname_path) / "bin").absolute())

    if (
        "PLATFORM_REPO_PATHS" not in os.environ.keys()
        or not Path(os.environ["PLATFORM_REPO_PATHS"]).exists()
    ):
        p = Path("/opt/xilinx/platforms")
        if p.exists():
            os.environ["PLATFORM_REPO_PATHS"] = str(p.absolute())
        else:
            warning(
                "PLATFORM_REPO_PATHS is not set "
                "and the default path does not exist. Synthesis might fail."
            )


def resolve_build_dir(
    flow_config: Path, build_dir: Path | None, settings: dict, is_test_run: bool = False
) -> Path | None:
    """Resolve the build dir. By default this should return FINN_TMP next to the flow config.
    Priority is command line argument > Environment variable > Settings Default > Fixed default.
    If the given path is relative and found in the envvar or settings file,
    add it to the flow_config_path"""
    if build_dir is not None:
        return build_dir
    if "FINN_BUILD_DIR" in os.environ.keys():
        p = Path(os.environ["FINN_BUILD_DIR"])
        if not p.is_absolute():
            return flow_config.parent / p
        return p
    if "FINN_BUILD_DIR" in settings.keys():
        p = Path(settings["FINN_BUILD_DIR"])
        if not p.is_absolute():
            return flow_config.parent / p
        return p
    if is_test_run:
        # Need a different fallback because tests have no build config
        return Path("/tmp/FINN_TEST_BUILD_DIR")
    return flow_config.parent / "FINN_TMP"


def resolve_deps_path(deps: Path | None, settings: dict) -> Path | None:
    """Try to resolve the dependency path. If none is valid, return None. If the path is relative,
    and not given by command line, append it to the FINN directory.
    Priority is command line argument > Environment variable > Settings Default > Fixed default"""
    if deps is not None:
        return deps
    if "FINN_DEPS" in os.environ.keys():
        p = Path(os.environ["FINN_DEPS"])
        if not p.is_absolute():
            return Path(__file__).parent.parent.parent.parent / p
        return p
    if "FINN_DEPS" in settings.keys():
        p = Path(settings["FINN_DEPS"])
        if not p.is_absolute():
            return Path(__file__).parent.parent.parent.parent / p
        return p
    return None


def resolve_num_workers(num: int, settings: dict) -> int:
    """Resolve the number of workers to use. Uses 75% of cores available as default fallback"""
    if num > -1:
        return num
    if "NUM_DEFAULT_WORKERS" in os.environ.keys() and os.environ["NUM_DEFAULT_WORKERS"] != "":
        return int(os.environ["NUM_DEFAULT_WORKERS"])
    if "NUM_DEFAULT_WORKERS" in settings.keys():
        return int(settings["NUM_DEFAULT_WORKERS"])
    cpus = multiprocessing.cpu_count()
    if cpus is None or cpus == 1:
        return 1
    return int(cpus * 0.75)


def read_yaml(p: Path) -> dict | None:
    """Read a yaml file and return its contents. If the file does not exist, return None"""
    if p.exists():
        with p.open() as f:
            return yaml.load(f, yaml.Loader)
    else:
        return None


def write_yaml(data: dict, p: Path) -> bool:
    """Try writing the given data to a yaml file. If this fails, return false otherwise
    true"""
    try:
        with p.open("w+") as f:
            yaml.dump(data, f, yaml.Dumper)
            return True
    except (OSError, yaml.error.YAMLError):
        return False



# Copyright (c) 2021, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import math
from onnx import TensorProto, helper
from qonnx.core.datatype import DataType
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.transformation.infer_datatypes import InferDataTypes
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import get_by_name


def _get_signed_from_upstream(model, trunc_node):
    """
    Find out what the sign of the input to the trunc node is,
    by looking at the upstream nodes.
    """
    node = trunc_node
    # Check if the input of this node already has a FINN datatype
    signed = None
    inp_dt = model.get_tensor_datatype(node.input[0])
    if inp_dt is not None and inp_dt != DataType["FLOAT32"]:
        signed = inp_dt.signed()
    # Go further up the graph, since the datatype inference works top down
    # these nodes should either be sign preserving ops or they already have a
    # datatype defined for the output tensor.
    curr_node = node
    if signed is None:
        while curr_node is not None:
            if model.is_join_node(curr_node):
                raise RuntimeError(
                    "Datatype Inference for the Trunc node only supports "
                    "linear nodes in the upstream path."
                )
            next_node = model.find_direct_predecessors(curr_node)
            if next_node is None:
                raise RuntimeError(
                    "Could not infere the Datatype for the Trunc node due to "
                    "missing upstream ndoes."
                )
            next_node = next_node[0]
            out_dt = model.get_tensor_datatype(next_node.output[0])
            if out_dt is not None and out_dt != DataType["FLOAT32"]:
                signed = out_dt.signed()
                break
            # Special cases where the node has an internal or intrinsic datatype.
            if next_node.op_type == "MultiThreshold":
                mt_inst = getCustomOp(next_node, onnx_opset_version=9)
                out_dt = DataType[mt_inst.get_nodeattr("out_dtype")]
                if out_dt is not None and out_dt != DataType["FLOAT32"]:
                    signed = out_dt.signed()
                    break
            if next_node.op_type == "BipolarQuant":
                signed = True
                break
            if next_node.op_type == "Quant":
                q_inst = getCustomOp(next_node, onnx_opset_version=9)
                out_dt = q_inst.get_integer_datatype(model)
                if out_dt is not None and out_dt != DataType["FLOAT32"]:
                    signed = out_dt.signed()
                    break

            # Check if we are allowed to move on to the next op
            sign_preserving_ops = ["Add", "Mul", "AveragePool", "Pad"]
            if next_node.op_type not in sign_preserving_ops:
                raise RuntimeError(
                    f"Could not infere the Datatype for the Trunc node, "
                    f"because the sign of the input datatype could not be infered "
                    f"from upstream nodes. And traversal further up the graph was "
                    f"disallowed, since the next node type {next_node.op_type} "
                    f"is not in the list of "
                    f"sign preserving ops {sign_preserving_ops}."
                )
            curr_node = next_node

    if signed is None:
        raise RuntimeError(
            "Could not infere the Datatype for the Trunc node, "
            "because the sign of the input datatype could not be infered "
            "from upstream nodes."
        )

    return signed


class AvgPoolAndTruncToQuantAvgPool(Transformation):
    """
    Convert a section of nodes of the pattern:
    AveragePool -> Mul (scalar) -> Trunc
    To the FINN op: QuantAvgPool2d
    """

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        for n in graph.node:
            node_ind += 1
            if n.op_type == "AveragePool":
                mul_node = model.find_direct_successors(n)
                if mul_node is not None and len(mul_node) == 1 and mul_node[0].op_type == "Mul":
                    mul_node = mul_node[0]
                    t_node = model.find_direct_successors(mul_node)
                    if t_node is not None and len(t_node) == 1 and t_node[0].op_type == "Trunc":
                        t_node = t_node[0]
                        running_node_index = node_ind
                        # Check node for compatibility
                        # Avg pooling node
                        k_s = get_by_name(n.attribute, "kernel_shape")
                        if k_s is None or len(k_s.ints) != 2 or len(set(k_s.ints)) != 1:
                            raise ValueError(
                                "FINN only supports average pooling with " "2D square kernels."
                            )
                        k_s = k_s.ints[0]

                        pads = get_by_name(n.attribute, "pads")
                        if pads is None or len(set(pads.ints)) != 1 or pads.ints[0] != 0:
                            raise ValueError("FINN dosn't support padding for average pooling.")

                        stride = get_by_name(n.attribute, "strides")
                        if stride is None or len(stride.ints) != 2 or len(set(stride.ints)) != 1:
                            raise ValueError(
                                "FINN only supports 2D strides with equal values in "
                                "each direction."
                            )
                        stride = stride.ints[0]

                        # Mul node
                        mul_val = model.get_initializer(mul_node.input[1])
                        if mul_val is None or len(mul_val.shape) != 0 or mul_val != k_s * k_s:
                            raise ValueError(
                                f"The Mul node after the AveragePool node must have "
                                f"static initialization at the second input, "
                                f"further the initialization must be of zero dimension "
                                f"and the value of the initialization must be "
                                f"the quadratic value of the kernel size, "
                                f"in this case {k_s * k_s}."
                            )

                        # Trunc node
                        rounding_mode = get_by_name(t_node.attribute, "rounding_mode")
                        normalized_mode_string = rounding_mode.s.upper()
                        if rounding_mode is None or normalized_mode_string != b"FLOOR":
                            raise ValueError(
                                "The Trunc node must have the rounding_mode " "set to 'FLOOR'."
                            )
                        for inp in t_node.input[1:]:
                            if model.get_initializer(inp) is None:
                                raise ValueError(
                                    f"All inputs of the Trunc node, "
                                    f"except the first, must be statically "
                                    f"initialized. However, {inp} is not."
                                )
                        zero_pt = model.get_initializer(t_node.input[2])
                        if len(zero_pt.shape) != 0 or zero_pt != 0:
                            raise ValueError(
                                f"Finn only supports 0 as the zero point for "
                                f"the Trunc node, it currently is {zero_pt}."
                            )
                        trunc_in_bits = model.get_initializer(t_node.input[3]).flatten()
                        trunc_out_bits = model.get_initializer(t_node.input[4]).flatten()
                        if len(trunc_in_bits.shape) != 1 or len(trunc_out_bits.shape) != 1:
                            raise ValueError(
                                f"Finn only supports scalar bit widths "
                                f"for the Trunc node. The input bit width "
                                f"currently is: {trunc_in_bits}, "
                                f"while the output bit width is: {trunc_out_bits}."
                            )
                        trunc_in_bits = int(trunc_in_bits[0])
                        trunc_out_bits = int(trunc_out_bits[0])

                        # Calculate parameters for the QuantAvgPool2d node,
                        # Calculate input bit width. Basically this backwards:
                        # https://github.com/Xilinx/finn-base/blob/
                        # 7c2603a95e90e4de2575020e575c24eab6a15889/src/finn/custom_op/
                        # general/quantavgpool2d.py#L94
                        ibits = math.floor(math.log(2**trunc_in_bits / (k_s * k_s), 2))
                        # Get sign
                        signed = _get_signed_from_upstream(model, t_node)
                        # ToDo: Change this to NHWC,
                        #  when the channels last layout comes around.
                        data_layout = "NCHW"

                        # Insert scale nodes, QuantAvgPool2d node and required tensors
                        scale = model.get_initializer(t_node.input[1])
                        scale_div_tensor = helper.make_tensor_value_info(
                            model.make_new_valueinfo_name(),
                            TensorProto.FLOAT,
                            None,
                        )
                        graph.value_info.append(scale_div_tensor)
                        model.set_initializer(scale_div_tensor.name, scale)

                        act_scale_div_tensor = helper.make_tensor_value_info(
                            model.make_new_valueinfo_name(),
                            TensorProto.FLOAT,
                            None,
                        )
                        graph.value_info.append(act_scale_div_tensor)

                        scale_div_node = helper.make_node(
                            "Div",
                            [n.input[0], scale_div_tensor.name],
                            [act_scale_div_tensor.name],
                        )
                        graph.node.insert(running_node_index, scale_div_node)
                        running_node_index += 1

                        act_scale_mul_tensor = helper.make_tensor_value_info(
                            model.make_new_valueinfo_name(),
                            TensorProto.FLOAT,
                            None,
                        )
                        graph.value_info.append(act_scale_mul_tensor)

                        QuantAvgPool2d_node = helper.make_node(
                            "QuantAvgPool2d",
                            [act_scale_div_tensor.name],
                            [act_scale_mul_tensor.name],
                            domain="qonnx.custom_op.general",
                            stride=stride,
                            kernel=k_s,
                            ibits=ibits,
                            obits=trunc_out_bits,
                            signed=int(signed),
                            data_layout=data_layout,
                        )
                        graph.node.insert(running_node_index, QuantAvgPool2d_node)
                        running_node_index += 1

                        scale_mul_tensor = helper.make_tensor_value_info(
                            model.make_new_valueinfo_name(),
                            TensorProto.FLOAT,
                            None,
                        )
                        graph.value_info.append(scale_mul_tensor)
                        model.set_initializer(scale_mul_tensor.name, scale)

                        scale_mul_node = helper.make_node(
                            "Mul",
                            [act_scale_mul_tensor.name, scale_mul_tensor.name],
                            [t_node.output[0]],
                        )
                        graph.node.insert(running_node_index, scale_mul_node)
                        running_node_index += 1

                        # Remove old nodes
                        graph.node.remove(n)
                        graph.node.remove(mul_node)
                        graph.node.remove(t_node)

                        # Recompute shapes and datatypes
                        model = model.transform(InferShapes())
                        model = model.transform(InferDataTypes())

                        return model, True

        return model, False



# Copyright (c) 2021, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import qonnx.core.onnx_exec as oxe
from onnx import TensorProto, helper
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.quant_constant_folding import FoldTransposeIntoQuantInit
from qonnx.transformation.remove import remove_node_and_rewire


class FoldQuantWeights(Transformation):
    """Merges Quant nodes, which are used as weights into the initializer
    of the weight tensor.
    """

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        graph_modified = False
        execution_context = model.make_empty_exec_context()
        for n in graph.node:
            node_ind += 1
            if n.op_type == "Quant" or n.op_type == "BipolarQuant":
                node_inp_inits = list(map(lambda x: model.get_initializer(x), n.input))
                node_inp_dyn = list(filter(lambda x: x is None, node_inp_inits))
                node_out = n.output[0]
                is_all_constant_inputs = len(node_inp_dyn) == 0
                ishape = model.get_tensor_shape(n.input[0])
                is_const_shape = (n.op_type == "Shape") and (ishape is not None)
                if is_all_constant_inputs or is_const_shape:
                    # Check node validity
                    if n.op_type == "Quant" and not model.get_initializer(n.input[2]) == 0:
                        raise ValueError(
                            "Only Quant nodes with zero-point == 0 " "are currently supported."
                        )
                    if model.is_fork_node(n):
                        raise ValueError(
                            "Weights quantized with the Quant node are not "
                            "allowed to be fork nodes node."
                        )
                    target_node = model.find_direct_successors(n)
                    if target_node is None:
                        raise RuntimeError(
                            "Weights quantized with the Quant node must have " "a successor node."
                        )
                    else:
                        target_node = target_node[0]
                    # If there is a DebugMarker in the weight path,
                    # then the DebugMarker needs to be removed before any further
                    # action is taken. Because this node interferes
                    # with how the constant folding tries to determine how to
                    # apply scale factors and in any case the DebugMarker would not
                    # return useful information after folding.
                    if target_node.op_type == "DebugMarker":
                        remove_node_and_rewire(model, target_node)
                        model = model.transform(FoldTransposeIntoQuantInit())
                        return model, True

                    # Continue with constant folding the quant node
                    scale = model.get_initializer(n.input[1])
                    unity_scale = (scale.flatten() == 1.0).all()
                    # this node has no dynamic inputs, only constant ones -- so we can
                    # do constant folding.
                    oxe.execute_node(n, execution_context, graph)
                    q_node_output = execution_context[node_out]
                    # Check we can directly constant fold
                    if unity_scale:
                        # use the execution result as an initializer
                        model.set_initializer(node_out, q_node_output)
                    else:
                        # Check next operator type
                        mul_like_nodes = [
                            "Mul",
                            "Div",
                            "Conv",
                            "MatMul",
                            "Gather",
                            "ConvTranspose",
                        ]
                        add_like_nodes = ["Add", "Sub"]
                        all_supported_ops = mul_like_nodes.copy()
                        all_supported_ops.extend(add_like_nodes)

                        if target_node.op_type not in all_supported_ops:
                            raise ValueError(
                                f"Can't constant fold Quant weight node "
                                f"into node type {target_node.op_type} "
                                f"at node: {target_node}."
                            )

                        # For both mul and Add:
                        # Move the scale factor behind the next operator
                        scale = model.get_initializer(n.input[1])
                        new_initializer = q_node_output / scale
                        # Round, to correct for floating point errors
                        new_initializer = np.round(new_initializer)
                        model.set_initializer(node_out, new_initializer)
                        q_inst = getCustomOp(n)
                        new_dtype = q_inst.get_integer_datatype(model)
                        model.set_tensor_datatype(node_out, new_dtype)

                        # Reshape scale for Conv if required
                        target_output_shape = model.get_tensor_shape(target_node.output[0])
                        if target_node.op_type == "Conv" and len(scale.shape) > 0:
                            conv_out_shape = [1] * len(target_output_shape)
                            # only support per-output channel scaling
                            # (i.e. all scale shape elems besides 0th must be 1s)
                            if len(scale.shape) > 1:
                                assert (
                                    np.prod(scale.shape[1:]) == 1
                                ), "Can't fold scale beyond per-out-channel granularity"
                            # collect all scaling in channels dim (since we constrain)
                            conv_out_shape[1] = -1
                            scale = scale.reshape(conv_out_shape)

                        if scale.shape == (1,):
                            scale = scale[0]
                            mul_shape = tuple()
                        else:
                            mul_shape = scale.shape
                        mul_tensor = helper.make_tensor_value_info(
                            model.make_new_valueinfo_name(),
                            TensorProto.FLOAT,
                            mul_shape,  # Note: This shape is known exactly as
                            # it is an initializer with known shape
                        )
                        graph.value_info.append(mul_tensor)
                        model.set_initializer(mul_tensor.name, scale)

                        successor = model.find_consumers(node_out)
                        if successor == []:
                            raise RuntimeError(
                                "Can only constant fold scaled Quant weights "
                                "if a successor exists."
                            )
                        assert len(successor) == 1, "Only implemented for a single consumer"
                        successor = successor[0]
                        succ_output_name = successor.output[0]

                        act_mul_tensor = helper.make_tensor_value_info(
                            model.make_new_valueinfo_name(),
                            TensorProto.FLOAT,
                            None,  # Note: Explicitly delete the shape
                            # annotation to be redone by the next shape
                            # inference
                        )
                        graph.value_info.append(act_mul_tensor)
                        successor.output[0] = act_mul_tensor.name

                        mul_node = helper.make_node(
                            "Mul",
                            [act_mul_tensor.name, mul_tensor.name],
                            [succ_output_name],
                        )
                        graph.node.insert(node_ind, mul_node)

                        if target_node.op_type in add_like_nodes:
                            # Move the scale factor behind also in-front of
                            # the next operator
                            div_tensor = helper.make_tensor_value_info(
                                model.make_new_valueinfo_name(),
                                TensorProto.FLOAT,
                                None,  # Note: Explicitly delete the shape
                                # annotation to be redone by the next shape
                                # inference
                            )
                            graph.value_info.append(div_tensor)
                            model.set_initializer(div_tensor.name, scale)

                            # Detect which input of the add-like successor is
                            # fed by the quantizer node to select the other
                            # branch to insert the scale factor
                            if successor.input[0] == node_out:
                                succ_input_name = successor.input[1]
                            else:
                                succ_input_name = successor.input[0]

                            act_mul_tensor = helper.make_tensor_value_info(
                                model.make_new_valueinfo_name(),
                                TensorProto.FLOAT,
                                None,  # Note: Explicitly delete the shape
                                # annotation to be redone by the next shape
                                # inference
                            )
                            graph.value_info.append(act_mul_tensor)

                            # Detect which input of the add-like successor is
                            # fed by the quantizer node to select the other
                            # branch to insert the scale factor
                            if successor.input[0] == node_out:
                                successor.input[1] = act_mul_tensor.name
                            else:
                                successor.input[0] = act_mul_tensor.name

                            div_node = helper.make_node(
                                "Div",
                                [succ_input_name, div_tensor.name],
                                [act_mul_tensor.name],
                            )
                            graph.node.insert(node_ind, div_node)

                    # remove old node
                    graph.node.remove(n)
                    graph_modified = True
                    # Note: Running shape inference is necessary as shape
                    # annotations have been deleted above
                    model = model.transform(InferShapes())
                    return (model, graph_modified)
        return (model, graph_modified)



"""
Base class for FINN benchmarking framework.

This module provides the foundational `bench` class for running automated benchmarks
of FINN dataflow builds. It wraps the existing FINN builder and handles configuration
management, build setup, and artifact collection.

Classes:
    bench: Main benchmarking class that orchestrates the complete build flow
"""

import glob
import os
import shutil
import yaml
from shutil import copy as shcopy
from shutil import copytree

import finn.builder.build_dataflow as build
import finn.builder.build_dataflow_config as build_cfg

from finn.benchmarking.util import delete_dir_contents
from finn.builder.build_dataflow_config import DataflowBuildConfig
from finn.util.basic import alveo_default_platform, alveo_part_map, part_map


class bench:
    """
    Base class for FINN benchmarking operations.

    This class provides the foundational framework for running automated benchmarks
    of FINN dataflow builds. It manages the complete lifecycle from configuration
    setup to build execution and artifact collection. Simple flows (also referred to
    as Designs-under-Tests (DUTs) in this context) can use this base class directly, while
    complex flows can subclass and override specific methods as needed.

    Attributes:
        report_dir (str): Directory where build reports are stored
        output_dict (dict): Collection of additional metrics produced by this infrastructure
    """

    def __init__(self, params, task_id, run_id, work_dir, artifacts_dir, save_dir, debug=True):
        """
        Initialize a new benchmark instance that manages a single FINN build.

        Args:
            params (dict): Parameters for the FINN builder and the bench instance itself
            task_id: Identifier for the SLURM task (job array index) or 0 if not using SLURM
            run_id: Unique identifier for this run instance
            work_dir (str): Working directory for build files
            artifacts_dir (str): Directory where output artifacts will be saved
            save_dir (str): Directory where additional (debug) artifacts will be saved
            debug (bool, optional): Enable debug mode for additional artifact collection.
                                  Defaults to True.

        The constructor performs several initialization tasks:
        - Sets up default board and timing configurations
        - Configures shell flow type based on board selection
        - Loads DUT-specific configuration from YAML files
        - Initializes artifact collection mechanisms
        - Prepares build directories and clears previous build artifacts
        """
        super().__init__()
        self._params = params
        self._task_id = task_id
        self._run_id = run_id
        self._work_dir = work_dir
        self._artifacts_dir = artifacts_dir
        self._save_dir = save_dir
        self._debug = debug

        # Setup some basic global default configuration
        # TODO: clean up or remove these attributes
        if "synth_clk_period_ns" in params:
            self._clock_period_ns = params["synth_clk_period_ns"]
        else:
            self._clock_period_ns = 10
            self._params["synth_clk_period_ns"] = self._clock_period_ns

        # TODO: do not allow multiple targets in a single bench job due to measurement?
        if "board" in params:
            self._board = params["board"]
        else:
            self._board = "RFSoC2x2"
            self._params["board"] = self._board

        if "part" in params:
            self._part = params["part"]
        elif self._board in part_map:
            self._part = part_map[self._board]
        else:
            raise Exception("No part specified for board %s" % self._board)

        if self._board in alveo_part_map:
            self._params["shell_flow_type"] = build_cfg.ShellFlowType.VITIS_ALVEO
            self._params["vitis_platform"] = alveo_default_platform[self._board]
        else:
            self._params["shell_flow_type"] = build_cfg.ShellFlowType.VIVADO_ZYNQ

        # Load custom (= non build_dataflow_config) parameters from topology-specific .yml
        custom_params = [
            "model_dir",  # used to setup onnx/npy input
            "model_path",  # used to setup onnx/npy input
            # model-gen parameters, such as seed, simd, pe, etc.
            # TODO: separate these more cleanly from builder options
        ]

        dut_yaml_name = self._params["dut"] + ".yml"
        dut_path = os.path.join(os.path.dirname(__file__), "dut", dut_yaml_name)
        if os.path.isfile(dut_path):
            with open(dut_path, "r") as f:
                dut_cfg = yaml.load(f, Loader=yaml.SafeLoader)
            for key in dut_cfg:
                if key in custom_params:
                    self._params[key] = dut_cfg[key]

        # Clear FINN tmp build dir before every run
        print("Clearing FINN BUILD DIR ahead of run")
        delete_dir_contents(os.environ["FINN_BUILD_DIR"])

        # Initialize dictionary to collect all benchmark results
        # TODO: remove completely or only use for meta data,
        # actual results go into run-specific .json files within /report
        self.output_dict = {}

        # Inputs (e.g., ONNX model, golden I/O pair, folding config, etc.)
        self._build_inputs = {}

        # Collect tuples of (name, source path, archive?) to save as pipeline artifacts
        self._artifacts_collection = []

        # Collect tuples of (name, source path, archive?) to save as local artifacts
        self._local_artifacts_collection = []
        if self._debug:
            # Save entire FINN_BUILD_DIR
            # TODO: add option to only save upon error/exception
            self._local_artifacts_collection.append(
                ("debug_finn_tmp", os.environ["FINN_BUILD_DIR"], True)
            )

        # SETUP
        # Use a temporary dir for buildflow-related files (next to FINN_BUILD_DIR)
        # Ensure it exists but is empty (clear potential artifacts from previous runs)
        tmp_buildflow_dir = os.path.join(self._work_dir, "buildflow")
        os.makedirs(tmp_buildflow_dir, exist_ok=True)
        delete_dir_contents(tmp_buildflow_dir)
        self._build_inputs["build_dir"] = os.path.join(
            tmp_buildflow_dir, "build_output"
        )  # TODO remove in favor of self.build_dir
        self._build_dir = os.path.join(tmp_buildflow_dir, "build_output")
        self.report_dir = os.path.join(self._build_dir, "report")
        os.makedirs(self.report_dir, exist_ok=True)

        # Save full build dir as local artifact
        self._local_artifacts_collection.append(("build_output", self._build_dir, False))
        # Save reports and deployment package as pipeline artifacts
        self._artifacts_collection.append(("reports", self.report_dir, False))
        self._artifacts_collection.append(
            ("reports", os.path.join(self._build_dir, "build_dataflow.log"), False)
        )
        self._artifacts_collection.append(("deploy", os.path.join(self._build_dir, "deploy"), True))

    def _save_artifact(self, target_path, source_path, archive=False):
        """
        Save a single artifact from source to target location.

        Args:
            target_path (str): Destination path where artifact will be saved
            source_path (str): Source path of the artifact to save
            archive (bool, optional): If True, create a ZIP archive of the source.
                                    Defaults to False.

        This method handles both files and directories:
        - For directories: copies recursively or creates ZIP archive
        - For files: copies to target directory
        - Automatically creates parent directories as needed
        """
        if os.path.isdir(source_path):
            if archive:
                os.makedirs(os.path.dirname(target_path), exist_ok=True)
                shutil.make_archive(target_path, "zip", source_path)
            else:
                os.makedirs(target_path, exist_ok=True)
                copytree(source_path, target_path, dirs_exist_ok=True)
        elif os.path.isfile(source_path):
            os.makedirs(target_path, exist_ok=True)
            shcopy(source_path, target_path)

    def save_artifacts_collection(self):
        """
        Save all collected pipeline artifacts.

        This method should be called upon successful or failed completion of a run.
        It processes all artifacts in the artifacts_collection list and saves them
        to the pipeline artifacts directory structure.

        Pipeline artifacts typically include:
        - Build reports and logs
        - Deployment packages
        """
        for name, source_path, archive in self._artifacts_collection:
            target_path = os.path.join(
                self._artifacts_dir, "runs_output", "run_%d" % (self._run_id), name
            )
            self._save_artifact(target_path, source_path, archive)

    def save_local_artifacts_collection(self):
        """
        Save all collected local artifacts for debugging.

        This method should be called upon successful or failed completion of a run.
        It processes all artifacts in the local_artifacts_collection list and saves
        them to the local save directory.

        Local artifacts typically include:
        - Complete build output directories
        - FINN build directory contents (when debug=True)
        """
        for name, source_path, archive in self._local_artifacts_collection:
            target_path = os.path.join(self._save_dir, name, "run_%d" % (self._run_id))
            self._save_artifact(target_path, source_path, archive)

    def _step_export_onnx(self):
        """
        Export or generate ONNX model for benchmarking.

        This method must be implemented by subclasses to provide the ONNX model
        that will be processed by the FINN build flow.

        Returns:
            str: Status indicator. Return "skipped" if no model can be generated
                 for the given parameters, which will cause the benchmark run
                 to be skipped.

        Note:
            This is an abstract method that must be overridden by concrete
            benchmark implementations.
        """
        pass

    def _step_build_setup(self):
        """
        Initialize the DataflowBuildConfig for this benchmark.

        This method can be overridden by subclasses if the setup is too complex
        for YAML definition. The default implementation loads configuration from
        a YAML file named after the DUT.

        Returns:
            DataflowBuildConfig: Initialized build configuration object

        Raises:
            Exception: If no DUT-specific YAML build definition is found

        The YAML file should be located at: benchmarking/dut/{dut_name}.yml
        where {dut_name} is the value of params["dut"].
        """
        dut_yaml_name = self._params["dut"] + ".yml"
        dut_path = os.path.join(os.path.dirname(__file__), "dut", dut_yaml_name)
        if os.path.isfile(dut_path):
            with open(dut_path, "r") as f:
                return DataflowBuildConfig.from_yaml(f)
        else:
            raise Exception("No DUT-specific YAML build definition found")

    def run(self):
        """
        Execute the benchmark run.

        This method defaults to running the complete FINN build flow but may be
        overridden by subclasses to implement custom benchmark sequences.

        Returns:
            str: Status of the benchmark run. Returns "skipped" if the benchmark
                 cannot be executed with the given parameters.
        """
        return self._steps_full_build_flow()

    def _step_parse_builder_output(self, build_dir):
        """
        Parse and analyze the output from the FINN builder.

        Args:
            build_dir (str): Path to the build output directory

        This method analyzes the build results and extracts verification status
        information. Currently focuses on checking verification step success by
        examining output files in the verification_output directory.

        The results are stored in self.output_dict for later analysis.

        TODO: Output results as .json or integrate as a new build step
        """
        if os.path.exists(os.path.join(build_dir, "verification_output")):
            # Collect all verification output filenames
            outputs = glob.glob(os.path.join(build_dir, "verification_output/*.npy"))
            # Extract the verification status for each verification output by matching
            # to the SUCCESS string contained in the filename
            status = all([out.split("_")[-1].split(".")[0] == "SUCCESS" for out in outputs])

            # Construct a dictionary reporting the verification status as string
            self.output_dict["builder_verification"] = {
                "verification": {True: "success", False: "fail"}[status]
            }
            # TODO: mark job as failed if verification fails?

    def _steps_full_build_flow(self):
        """
        Execute the complete FINN dataflow build sequence.

        This method implements the default step sequence for benchmarking a full
        FINN builder flow, including:

        1. **Model Creation/Import**: Load or generate ONNX model
        2. **Build Setup**: Configure DataflowBuildConfig with parameters
        3. **Build Execution**: Run the FINN dataflow build pipeline
        4. **Analysis**: Parse and collect build results

        Returns:
            str: "skipped" if the benchmark cannot be executed, otherwise None

        The method handles three model input scenarios:
        - model_dir: Pre-existing ONNX model with verification I/O pairs
        - model_path: Path to existing ONNX model file
        - Generated: ONNX model created by _step_export_onnx()

        Configuration management includes:
        - Loading base configuration from DUT YAML file
        - Setting global defaults for optimization and debugging
        - Merging run-specific parameters
        - Environment setup for build execution
        """
        if "model_dir" in self._params:
            # input ONNX model and verification input/output pairs are provided
            model_dir = self._params["model_dir"]
            self._build_inputs["onnx_path"] = os.path.join(model_dir, "model.onnx")
            self._build_inputs["input_npy_path"] = os.path.join(model_dir, "inp.npy")
            self._build_inputs["output_npy_path"] = os.path.join(model_dir, "out.npy")
        elif "model_path" in self._params:
            self._build_inputs["onnx_path"] = self._params["model_path"]
        else:
            # input ONNX model (+ optional I/O pair for verification) will be generated
            self._build_inputs["onnx_path"] = os.path.join(
                self._build_inputs["build_dir"], "model_export.onnx"
            )
            if self._step_export_onnx(self._build_inputs["onnx_path"]) == "skipped":
                # microbenchmarks might skip because no model can be generated for given params
                return "skipped"

        # BUILD SETUP
        # Initialize from YAML (default) or custom script (if dedicated subclass is defined)
        cfg = self._step_build_setup()

        # Set some global defaults (could still be overwritten by run-specific YAML)
        cfg.output_dir = self._build_inputs["build_dir"]
        # enable extra performance optimizations (physopt)
        # TODO: check OMX synth strategy again!
        cfg.vitis_opt_strategy = build_cfg.VitisOptStrategy.PERFORMANCE_BEST
        cfg.verbose = True
        cfg.console_log_level = build_cfg.LogLevel.ERROR
        cfg.enable_build_pdb_debug = False
        cfg.enable_exception_snapshots = True
        # cfg.stitched_ip_gen_dcp = False # only needed for further manual integration
        cfg.split_large_fifos = True
        cfg.save_intermediate_models = True  # Save the intermediate model graphs
        cfg.verify_save_full_context = True  # Output full context dump for verification steps
        cfg.enable_instrumentation = True
        # rtlsim_use_vivado_comps # TODO ?
        # cfg.default_swg_exception
        # cfg.large_fifo_mem_style

        # Overwrite build config settings with run-specific parameters
        # Filter to only valid DataflowBuildConfig attributes to avoid errors
        valid_params = {k: v for k, v in self._params.items() if hasattr(cfg, k)}

        # TODO: warn/error if there are unrecognized options set?

        if valid_params:
            # Use DataflowBuildConfig's from_dict method which handles enum conversion.
            updated_cfg = DataflowBuildConfig.from_dict(valid_params)

            # Only apply values that were explicitly specified in params (not defaults)
            for param_key in valid_params.keys():
                setattr(cfg, param_key, getattr(updated_cfg, param_key))

        # Default of 1M cycles is insufficient for MetaFi (6M) and RN-50 (2.5M)
        # TODO: make configurable or set on pipeline level?
        os.environ["LIVENESS_THRESHOLD"] = "10000000"

        # BUILD
        build.build_dataflow_cfg(self._build_inputs["onnx_path"], cfg)

        # ANALYSIS
        self._step_parse_builder_output(self._build_inputs["build_dir"])



# Copyright (c) 2020, Xilinx, Inc.
# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Transformation to create stitched IP from dataflow graph components."""

import json
import multiprocessing as mp
import os
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.util.basic import get_num_default_workers
from shutil import copytree
from subprocess import CalledProcessError

from finn.templates import get_templates_folder
from finn.transformation.fpgadataflow.replace_verilog_relpaths import ReplaceVerilogRelPaths
from finn.util.basic import launch_process_helper, make_build_dir
from finn.util.exception import FINNError, FINNUserError
from finn.util.fpgadataflow import is_hls_node, is_rtl_node
from finn.util.logging import log


def is_external_input(model, node, i):
    """
    Determine whether input i of node should be made external.

    True only if input is unconnected and has no initializer.
    Only exception is second input of FC layers when mem_mode is external.
    """
    node_inst = getCustomOp(node)
    op_type = node.op_type
    producer = model.find_producer(node.input[i])
    if producer is None:
        if model.get_initializer(node.input[i]) is None:
            return True
        else:
            if op_type.startswith("MVAU"):
                if node_inst.get_nodeattr("mem_mode") == "external":
                    return True
    return False


def is_external_output(model, node, i):
    """Determine whether output i of node should be made external."""
    consumers = model.find_consumers(node.output[i])
    if consumers == []:
        # TODO should ideally check if tensor is in top-level
        # outputs
        return True
    return False


class CreateStitchedIP(Transformation):
    """Create a Vivado IP Block Design project from all the generated IPs of a
    graph. All nodes in the graph must have the fpgadataflow backend attribute,
    and the PrepareIP transformation must have been previously run on
    the graph. The resulting block design is also packaged as IP. The
    transformation gets the fpgapart as a string.

    Outcome if successful: sets the vivado_stitch_proj attribute in the ONNX
    ModelProto's metadata_props field, with the created project dir as the
    value. A make_project.tcl script is also placed under the same folder,
    which is called to instantiate the per-layer IPs and stitch them together.
    The packaged block design IP can be found under the ip subdirectory.
    """

    def __init__(self, fpgapart, clk_ns, ip_name="finn_design", vitis=False, signature=[]):
        """Initialize CreateStitchedIP transformation with FPGA part and clock settings."""
        super().__init__()
        self.fpgapart = fpgapart
        self.clk_ns = clk_ns
        self.ip_name = ip_name
        self.vitis = vitis
        self.signature = signature
        self.has_aximm = False
        self.has_m_axis = False
        self.m_axis_idx = 0
        self.has_s_axis = False
        self.s_axis_idx = 0
        self.clock_reset_are_external = False
        self.clock2x_is_external = False
        self.create_cmds = []
        self.connect_cmds = []
        # keep track of top-level interface names
        self.intf_names = {
            "clk": [],
            "rst": [],
            "s_axis": [],
            "m_axis": [],
            "aximm": [],
            "axilite": [],
        }

    def is_double_pumped(self, node):
        """Check if node uses double pumped computation."""
        if node.op_type.startswith("MVAU"):
            inst = getCustomOp(node)
            try:
                pumped_compute = inst.get_nodeattr("pumpedCompute")
            except AttributeError:
                pumped_compute = 0
            return pumped_compute or inst.get_nodeattr("pumpedMemory")

    def connect_clk_rst(self, node):
        """Connect clock and reset signals for the node."""
        inst_name = node.name
        node_inst = getCustomOp(node)
        clock_intf_name = node_inst.get_verilog_top_module_intf_names()["clk"][0]
        reset_intf_name = node_inst.get_verilog_top_module_intf_names()["rst"][0]
        # make clock and reset external, if they aren't already
        if not self.clock_reset_are_external:
            self.connect_cmds.append(
                "make_bd_pins_external [get_bd_pins %s/%s]" % (inst_name, clock_intf_name)
            )
            self.connect_cmds.append("set_property name ap_clk [get_bd_ports ap_clk_0]")
            self.connect_cmds.append(
                "make_bd_pins_external [get_bd_pins %s/%s]" % (inst_name, reset_intf_name)
            )
            self.connect_cmds.append("set_property name ap_rst_n [get_bd_ports ap_rst_n_0]")
            self.clock_reset_are_external = True
            self.intf_names["clk"] = ["ap_clk"]
            self.intf_names["rst"] = ["ap_rst_n"]
        # otherwise connect clock and reset
        else:
            self.connect_cmds.append(
                "connect_bd_net [get_bd_ports ap_rst_n] [get_bd_pins %s/%s]"
                % (inst_name, reset_intf_name)
            )
            self.connect_cmds.append(
                "connect_bd_net [get_bd_ports ap_clk] [get_bd_pins %s/%s]"
                % (inst_name, clock_intf_name)
            )
        # make clk2x external, if it isn't already and connect clk2x
        if self.is_double_pumped(node):
            clock2x_intf_name = node_inst.get_verilog_top_module_intf_names()["clk2x"][0]
            if not self.clock2x_is_external:
                self.connect_cmds.append(
                    "make_bd_pins_external [get_bd_pins %s/%s]" % (inst_name, clock2x_intf_name)
                )
                self.connect_cmds.append("set_property name ap_clk2x [get_bd_ports ap_clk2x_0]")
                self.clock2x_is_external = True
                self.intf_names["clk2x"] = ["ap_clk2x"]
            # otherwise connect clk2x
            else:
                if self.is_double_pumped(node):
                    self.connect_cmds.append(
                        "connect_bd_net [get_bd_ports ap_clk2x] [get_bd_pins %s/%s]"
                        % (inst_name, clock2x_intf_name)
                    )

    def connect_axi(self, node):
        """Connect AXI interfaces for the node."""
        inst_name = node.name
        node_inst = getCustomOp(node)
        axilite_intf_name = node_inst.get_verilog_top_module_intf_names()["axilite"]
        aximm_intf_name = node_inst.get_verilog_top_module_intf_names()["aximm"]
        if len(axilite_intf_name) != 0:
            self.connect_cmds.append(
                "make_bd_intf_pins_external "
                "[get_bd_intf_pins %s/%s]" % (inst_name, axilite_intf_name[0])
            )
            ext_if_name = "%s_%d" % (
                axilite_intf_name[0],
                len(self.intf_names["axilite"]),
            )
            self.intf_names["axilite"].append(ext_if_name)
        if len(aximm_intf_name) != 0:
            self.connect_cmds.append(
                "make_bd_intf_pins_external [get_bd_intf_pins %s/%s]"
                % (inst_name, aximm_intf_name[0][0])
            )
            ext_if_name = "m_axi_gmem%d" % (len(self.intf_names["aximm"]))
            self.connect_cmds.append(
                "set_property name %s [get_bd_intf_ports m_axi_gmem_0]" % ext_if_name
            )
            self.connect_cmds.append("assign_bd_address")
            seg_name = "%s/Data_m_axi_gmem/SEG_%s_Reg" % (inst_name, ext_if_name)
            self.connect_cmds.append("set_property offset 0 [get_bd_addr_segs {%s}]" % (seg_name))
            # TODO should propagate this information from the node instead of 4G
            self.connect_cmds.append("set_property range 4G [get_bd_addr_segs {%s}]" % (seg_name))
            self.intf_names["aximm"] = [(ext_if_name, aximm_intf_name[0][1])]
            self.has_aximm = True

    def connect_m_axis_external(self, node, idx=None):
        """Connect master AXI stream interfaces as external ports."""
        inst_name = node.name
        node_inst = getCustomOp(node)
        output_intf_names = node_inst.get_verilog_top_module_intf_names()["m_axis"]
        # make output axis external
        for i in range(len(output_intf_names)):
            if idx is not None and idx != i:
                continue
            output_intf_name = output_intf_names[i][0]
            self.connect_cmds.append(
                "make_bd_intf_pins_external [get_bd_intf_pins %s/%s]"
                % (inst_name, output_intf_name)
            )
            self.connect_cmds.append(
                "set_property name m_axis_%d [get_bd_intf_ports %s_0]"
                % (self.m_axis_idx, output_intf_name)
            )
            self.has_m_axis = True
            self.intf_names["m_axis"].append(
                ("m_axis_%d" % self.m_axis_idx, output_intf_names[i][1])
            )
            self.m_axis_idx += 1

    def connect_s_axis_external(self, node, idx=None):
        """Connect slave AXI stream interfaces as external ports."""
        inst_name = node.name
        node_inst = getCustomOp(node)
        input_intf_names = node_inst.get_verilog_top_module_intf_names()["s_axis"]
        # make input axis external
        for i in range(len(input_intf_names)):
            if idx is not None and idx != i:
                continue
            input_intf_name = input_intf_names[i][0]
            self.connect_cmds.append(
                "make_bd_intf_pins_external [get_bd_intf_pins %s/%s]" % (inst_name, input_intf_name)
            )
            self.connect_cmds.append(
                "set_property name s_axis_%d [get_bd_intf_ports %s_0]"
                % (self.s_axis_idx, input_intf_name)
            )
            self.has_s_axis = True
            self.intf_names["s_axis"].append(
                ("s_axis_%d" % self.s_axis_idx, input_intf_names[i][1])
            )
            self.s_axis_idx += 1

    def connect_ap_none_external(self, node):
        """Connect ap_none interfaces as external ports."""
        inst_name = node.name
        node_inst = getCustomOp(node)
        input_intf_names = node_inst.get_verilog_top_module_intf_names()["ap_none"]
        # make external
        for i in range(len(input_intf_names)):
            input_intf_name = input_intf_names[i]
            self.connect_cmds.append(
                "make_bd_pins_external [get_bd_pins %s/%s]" % (inst_name, input_intf_name)
            )
            self.connect_cmds.append(
                "set_property name %s [get_bd_ports %s_0]" % (input_intf_name, input_intf_name)
            )

    def insert_signature(self, checksum_count):
        """Insert signature block for design identification."""
        signature_vlnv = "AMD:user:axi_info_top:1.0"
        signature_name = "axi_info_top0"
        self.create_cmds.append(
            "create_bd_cell -type ip -vlnv %s %s" % (signature_vlnv, signature_name)
        )
        self.create_cmds.append(
            "set_property -dict [list "
            "CONFIG.SIG_CUSTOMER {%s} "
            "CONFIG.SIG_APPLICATION {%s} "
            "CONFIG.VERSION {%s} "
            "CONFIG.CHECKSUM_COUNT {%s} "
            "] [get_bd_cells %s]"
            % (
                self.signature[0],
                self.signature[1],
                self.signature[2],
                checksum_count,
                signature_name,
            )
        )
        # set clk and reset
        self.connect_cmds.append(
            "connect_bd_net [get_bd_ports ap_clk] [get_bd_pins %s/ap_clk]" % signature_name
        )
        self.connect_cmds.append(
            "connect_bd_net [get_bd_ports ap_rst_n] [get_bd_pins %s/ap_rst_n]" % signature_name
        )
        fclk_mhz = 1 / (self.clk_ns * 0.001)
        fclk_hz = fclk_mhz * 1000000
        self.connect_cmds.append(
            "set_property -dict [list "
            "CONFIG.FREQ_HZ {%f} "
            "CONFIG.CLK_DOMAIN {ap_clk} "
            "] [get_bd_intf_pins %s/s_axi]"
            % (
                fclk_hz,
                signature_name,
            )
        )
        # make axilite interface external
        self.connect_cmds.append(
            "make_bd_intf_pins_external [get_bd_intf_pins %s/s_axi]" % signature_name
        )
        self.connect_cmds.append("set_property name s_axilite_info [get_bd_intf_ports s_axi_0]")
        self.connect_cmds.append("assign_bd_address")

    def apply(self, model):
        """Apply the CreateStitchedIP transformation to the model."""
        # ensure non-relative readmemh .dat files
        model = model.transform(ReplaceVerilogRelPaths())
        ip_dirs = ["list"]
        # add RTL streamer IP
        ip_dirs.append("$::env(FINN_RTLLIB)/memstream")
        if self.signature:
            ip_dirs.append("$::env(FINN_RTLLIB)/axi_info")
        if model.graph.node[0].op_type not in ["StreamingFIFO_rtl", "IODMA_hls"]:
            log.warning(
                """First node is not StreamingFIFO or IODMA.
                You may experience incorrect stitched-IP rtlsim or hardware
                behavior. It is strongly recommended to insert FIFOs prior to
                calling CreateStitchedIP."""
            )
        if model.graph.node[0].op_type == "StreamingFIFO_rtl":
            firstfifo = getCustomOp(model.graph.node[0])
            if firstfifo.get_nodeattr("impl_style") == "vivado":
                log.warning(
                    """First FIFO has impl_style=vivado, which may cause
                    simulation glitches (e.g. dropping the first input sample
                    after reset)."""
                )
        for node in model.graph.node:
            # ensure that all nodes are fpgadataflow, and that IPs are generated
            assert is_hls_node(node) or is_rtl_node(
                node
            ), "All nodes must be FINN fpgadataflow nodes."
            node_inst = getCustomOp(node)
            ip_dir_value = node_inst.get_nodeattr("ip_path")
            assert os.path.isdir(ip_dir_value), "IP generation directory doesn't exist."
            ip_dirs += [ip_dir_value]
            self.create_cmds += node_inst.code_generation_ipi()
            self.connect_clk_rst(node)
            self.connect_ap_none_external(node)
            self.connect_axi(node)
            for i in range(len(node.input)):
                if not is_external_input(model, node, i):
                    producer = model.find_producer(node.input[i])
                    if producer is None:
                        continue
                    j = list(producer.output).index(node.input[i])
                    src_intf_name = getCustomOp(producer).get_verilog_top_module_intf_names()[
                        "m_axis"
                    ][j][0]
                    dst_intf_name = node_inst.get_verilog_top_module_intf_names()["s_axis"][i][0]
                    self.connect_cmds.append(
                        "connect_bd_intf_net [get_bd_intf_pins %s/%s] "
                        "[get_bd_intf_pins %s/%s]"
                        % (producer.name, src_intf_name, node.name, dst_intf_name)
                    )

        # process external inputs and outputs in top-level graph input order
        for input in model.graph.input:
            inp_name = input.name
            inp_cons = model.find_consumers(inp_name)
            assert inp_cons != [], "No consumer for input " + inp_name
            assert len(inp_cons) == 1, "Multiple consumers for input " + inp_name
            node = inp_cons[0]
            node_inst = getCustomOp(node)
            for i in range(len(node.input)):
                if node.input[i] == inp_name:
                    self.connect_s_axis_external(node, idx=i)
        for output in model.graph.output:
            out_name = output.name
            node = model.find_producer(out_name)
            assert node is not None, "No producer for output " + out_name
            node_inst = getCustomOp(node)
            for i in range(len(node.output)):
                if node.output[i] == out_name:
                    self.connect_m_axis_external(node, idx=i)

        if self.signature:
            # extract number of checksum layer from graph
            checksum_layers = model.get_nodes_by_op_type("CheckSum_hls")
            self.insert_signature(len(checksum_layers))

        # create a temporary folder for the project
        prjname = "finn_vivado_stitch_proj"
        vivado_stitch_proj_dir = make_build_dir(prefix="vivado_stitch_proj_")
        model.set_metadata_prop("vivado_stitch_proj", vivado_stitch_proj_dir)
        # start building the tcl script
        tcl = []
        # create vivado project
        tcl.append(
            "create_project %s %s -part %s" % (prjname, vivado_stitch_proj_dir, self.fpgapart)
        )
        # no warnings on long module names
        tcl.append("set_msg_config -id {[BD 41-1753]} -suppress")
        # add all the generated IP dirs to ip_repo_paths
        ip_dirs_str = " ".join(ip_dirs)
        tcl.append("set_property ip_repo_paths [%s] [current_project]" % ip_dirs_str)
        tcl.append("update_ip_catalog")
        # create block design and instantiate all layers
        block_name = self.ip_name
        tcl.append('create_bd_design "%s"' % block_name)
        tcl.extend(self.create_cmds)
        tcl.extend(self.connect_cmds)
        fclk_mhz = 1 / (self.clk_ns * 0.001)
        fclk_hz = fclk_mhz * 1000000
        tcl.append("set_property CONFIG.FREQ_HZ %d [get_bd_ports /ap_clk]" % round(fclk_hz))
        if self.clock2x_is_external:
            tcl.append(
                "set_property CONFIG.FREQ_HZ %d [get_bd_ports /ap_clk2x]" % round(2 * fclk_hz)
            )
        tcl.append("validate_bd_design")
        tcl.append("save_bd_design")
        # create wrapper hdl (for rtlsim later on)
        bd_base = "%s/%s.srcs/sources_1/bd/%s" % (
            vivado_stitch_proj_dir,
            prjname,
            block_name,
        )
        bd_filename = "%s/%s.bd" % (bd_base, block_name)
        tcl.append("make_wrapper -files [get_files %s] -top" % bd_filename)
        wrapper_filename = "%s/hdl/%s_wrapper.v" % (bd_base, block_name)
        tcl.append("add_files -norecurse %s" % wrapper_filename)
        model.set_metadata_prop("wrapper_filename", wrapper_filename)
        tcl.append("set_property top %s_wrapper [current_fileset]" % block_name)
        # synthesize to DCP and export stub, DCP and constraints
        if self.vitis:
            tcl.append(
                "set_property SYNTH_CHECKPOINT_MODE Hierarchical [ get_files %s ]" % bd_filename
            )
            tcl.append(
                "set_property -name {STEPS.SYNTH_DESIGN.ARGS.MORE OPTIONS} "
                "-value {-mode out_of_context} -objects [get_runs synth_1]"
            )
            num_workers = get_num_default_workers()
            assert num_workers >= 0, "Number of workers must be nonnegative."
            if num_workers == 0:
                num_workers = mp.cpu_count()
            tcl.append("launch_runs synth_1 -jobs %s" % str(num_workers))
            tcl.append("wait_on_run [get_runs synth_1]")
            tcl.append("open_run synth_1 -name synth_1")
            tcl.append("write_verilog -force -mode synth_stub %s.v" % block_name)
            tcl.append("write_checkpoint %s.dcp" % block_name)
            tcl.append("write_xdc %s.xdc" % block_name)
            tcl.append(
                "report_utilization -hierarchical -hierarchical_depth 5 "
                "-file %s_partition_util.rpt" % block_name
            )
        # export block design itself as an IP core
        block_vendor = "xilinx_finn"
        block_library = "finn"
        block_vlnv = "%s:%s:%s:1.0" % (block_vendor, block_library, block_name)
        model.set_metadata_prop("vivado_stitch_vlnv", block_vlnv)
        model.set_metadata_prop("vivado_stitch_ifnames", json.dumps(self.intf_names))
        tcl.append(
            (
                "ipx::package_project -root_dir %s/ip -vendor %s "
                "-library %s -taxonomy /UserIP -module %s -import_files"
            )
            % (vivado_stitch_proj_dir, block_vendor, block_library, block_name)
        )
        # Allow user to customize clock in deployment of stitched IP
        tcl.append("set_property ipi_drc {ignore_freq_hz true} [ipx::current_core]")
        # in some cases, the IP packager seems to infer an aperture of 64K or 4G,
        # preventing address assignment of the DDR_LOW and/or DDR_HIGH segments
        # the following is a hotfix to remove this aperture during IODMA packaging
        tcl.append(
            "ipx::remove_segment -quiet m_axi_gmem0:APERTURE_0 "
            "[ipx::get_address_spaces m_axi_gmem0 -of_objects [ipx::current_core]]"
        )
        tcl.append("set_property core_revision 2 [ipx::find_open_core %s]" % block_vlnv)
        tcl.append("ipx::create_xgui_files [ipx::find_open_core %s]" % block_vlnv)
        # mark bus interface params as user-resolvable to avoid FREQ_MHZ mismatches
        tcl.append(
            "set_property value_resolve_type user [ipx::get_bus_parameters "
            "-of [ipx::get_bus_interfaces -of [ipx::current_core ]]]"
        )
        # if targeting Vitis, add some properties to the IP
        if self.vitis:
            # replace source code with dcp
            tcl.append("set_property sdx_kernel true [ipx::find_open_core %s]" % block_vlnv)
            tcl.append("set_property sdx_kernel_type rtl [ipx::find_open_core %s]" % block_vlnv)
            tcl.append("set_property supported_families { } [ipx::find_open_core %s]" % block_vlnv)
            tcl.append(
                "set_property xpm_libraries {XPM_CDC XPM_MEMORY XPM_FIFO} "
                "[ipx::find_open_core %s]" % block_vlnv
            )
            tcl.append(
                "set_property auto_family_support_level level_2 "
                "[ipx::find_open_core %s]" % block_vlnv
            )
            # remove all files from synthesis and sim groups
            # we'll replace with DCP, stub, and xdc
            tcl.append(
                "ipx::remove_all_file "
                "[ipx::get_file_groups xilinx_anylanguagebehavioralsimulation]"
            )
            tcl.append("ipx::remove_all_file " "[ipx::get_file_groups xilinx_anylanguagesynthesis]")
            tcl.append(
                "ipx::remove_file_group "
                "xilinx_anylanguagebehavioralsimulation [ipx::current_core]"
            )
            tcl.append("ipx::remove_file_group " "xilinx_anylanguagesynthesis [ipx::current_core]")
            # remove sim and src folders
            tcl.append("file delete -force %s/ip/sim" % vivado_stitch_proj_dir)
            tcl.append("file delete -force %s/ip/src" % vivado_stitch_proj_dir)
            # copy and add DCP, stub, and xdc
            tcl.append("file mkdir %s/ip/dcp" % vivado_stitch_proj_dir)
            tcl.append("file mkdir %s/ip/impl" % vivado_stitch_proj_dir)
            tcl.append("file copy -force %s.dcp %s/ip/dcp" % (block_name, vivado_stitch_proj_dir))
            tcl.append("file copy -force %s.xdc %s/ip/impl" % (block_name, vivado_stitch_proj_dir))
            tcl.append("ipx::add_file_group xilinx_implementation [ipx::current_core]")
            tcl.append(
                "ipx::add_file impl/%s.xdc [ipx::get_file_groups xilinx_implementation]"
                % block_name
            )
            tcl.append(
                "set_property used_in [list implementation] "
                "[ipx::get_files impl/%s.xdc "
                "-of_objects [ipx::get_file_groups xilinx_implementation]]" % block_name
            )
            tcl.append("ipx::add_file_group " "xilinx_synthesischeckpoint [ipx::current_core]")
            tcl.append(
                "ipx::add_file dcp/%s.dcp "
                "[ipx::get_file_groups xilinx_synthesischeckpoint]" % block_name
            )
            tcl.append("ipx::add_file_group xilinx_simulationcheckpoint [ipx::current_core]")
            tcl.append(
                "ipx::add_file dcp/%s.dcp "
                "[ipx::get_file_groups xilinx_simulationcheckpoint]" % block_name
            )
        # add a rudimentary driver mdd to get correct ranges in xparameters.h later on
        min_driver = get_templates_folder() / "ipcore_driver"
        copytree(min_driver, vivado_stitch_proj_dir + "/data")

        #####
        # Core Cleanup Operations
        tcl.append(
            """
set core [ipx::current_core]

# Add rudimentary driver
file copy -force data ip/
set file_group [ipx::add_file_group -type software_driver {} $core]
set_property type mdd       [ipx::add_file data/finn_design.mdd $file_group]
set_property type tclSource [ipx::add_file data/finn_design.tcl $file_group]

# Remove all XCI references to subcores
set impl_files [ipx::get_file_groups xilinx_implementation -of $core]
foreach xci [ipx::get_files -of $impl_files {*.xci}] {
    ipx::remove_file [get_property NAME $xci] $impl_files
}

# Construct a single flat memory map for each AXI-lite interface port
foreach port [get_bd_intf_ports -filter {CONFIG.PROTOCOL==AXI4LITE}] {
    set pin $port
    set awidth ""
    while { $awidth == "" } {
        set pins [get_bd_intf_pins -of [get_bd_intf_nets -boundary_type lower -of $pin]]
        set kill [lsearch $pins $pin]
        if { $kill >= 0 } { set pins [lreplace $pins $kill $kill] }
        if { [llength $pins] != 1 } { break }
        set pin [lindex $pins 0]
        set awidth [get_property CONFIG.ADDR_WIDTH $pin]
    }
    if { $awidth == "" } {
       puts "CRITICAL WARNING: Unable to construct address map for $port."
    } {
       set range [expr 2**$awidth]
       set range [expr $range < 4096 ? 4096 : $range]
       puts "INFO: Building address map for $port: 0+:$range"
       set name [get_property NAME $port]
       set addr_block [ipx::add_address_block Reg0 [ipx::add_memory_map $name $core]]
       set_property range $range $addr_block
       set_property slave_memory_map_ref $name [ipx::get_bus_interfaces $name -of $core]
    }
}

# Finalize and Save
ipx::update_checksums $core
ipx::save_core $core

# Remove stale subcore references from component.xml
file rename -force ip/component.xml ip/component.bak
set ifile [open ip/component.bak r]
set ofile [open ip/component.xml w]
set buf [list]
set kill 0
while { [eof $ifile] != 1 } {
    gets $ifile line
    if { [string match {*<spirit:fileSet>*} $line] == 1 } {
        foreach l $buf { puts $ofile $l }
        set buf [list $line]
    } elseif { [llength $buf] > 0 } {
        lappend buf $line

        if { [string match {*</spirit:fileSet>*} $line] == 1 } {
            if { $kill == 0 } { foreach l $buf { puts $ofile $l } }
            set buf [list]
            set kill 0
        } elseif { [string match {*<xilinx:subCoreRef>*} $line] == 1 } {
            set kill 1
        }
    } else {
        puts $ofile $line
    }
}
close $ifile
close $ofile
"""
        )

        # export list of used Verilog files (for rtlsim later on)
        tcl.append(
            "set all_v_files [get_files -filter {USED_IN_SYNTHESIS == 1 "
            + "&& (FILE_TYPE == Verilog || FILE_TYPE == SystemVerilog "
            + '|| FILE_TYPE =="Verilog Header")}]'
        )
        v_file_list = "%s/all_verilog_srcs.txt" % vivado_stitch_proj_dir
        tcl.append("set fp [open %s w]" % v_file_list)
        # write each verilog filename to all_verilog_srcs.txt
        tcl.append("foreach vf $all_v_files {puts $fp $vf}")
        tcl.append("close $fp")
        # write the project creator tcl script
        tcl_string = "\n".join(tcl) + "\n"
        with open(vivado_stitch_proj_dir + "/make_project.tcl", "w") as f:
            f.write(tcl_string)
        # create a shell script and call Vivado
        make_project_sh = vivado_stitch_proj_dir + "/make_project.sh"
        working_dir = os.getcwd()
        with open(make_project_sh, "w") as f:
            f.write("#!/bin/bash \n")
            f.write("cd {}\n".format(vivado_stitch_proj_dir))
            f.write("set -e\n")  # Exit with non-zero if vivado fails.
            f.write("vivado -mode batch -source make_project.tcl\n")
            f.write("cd {}\n".format(working_dir))
        bash_command = ["bash", make_project_sh]

        try:
            launch_process_helper(bash_command, print_stdout=False)
        except CalledProcessError as e:
            raise FINNUserError(
                f"CreateStitchedIP: make_project.sh failed with a non-zero "
                f"exit code. Check previous logs and logs in "
                f"{vivado_stitch_proj_dir} to find out why it failed."
            ) from e

        # wrapper may be created in different location depending on Vivado version
        if not os.path.isfile(wrapper_filename):
            # check in alternative location (.gen instead of .srcs)
            wrapper_filename_alt = wrapper_filename.replace(".srcs", ".gen")
            if os.path.isfile(wrapper_filename_alt):
                model.set_metadata_prop("wrapper_filename", wrapper_filename_alt)
            else:
                raise FINNError(
                    """CreateStitchedIP failed, no wrapper HDL found under %s or %s.
                    Please check logs under the parent directory."""
                    % (wrapper_filename, wrapper_filename_alt)
                )

        return (model, False)



# Copyright (C) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnxruntime as rt
from onnx import TensorProto, helper
from qonnx.core.datatype import DataType
from qonnx.util.basic import qonnx_make_model

from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp
from finn.util.logging import log

# ONNX i/o tensor shape assumptions for channelwise ops:
# input 0 is the input tensor, shape (..., NumChannels)
# input 1 is the channelwise parameter tensor, shape (NumChannels, params_per_channel)
# output 0 is the output tensor, shape (..., NumChannels) - same as input
# the ... here can be any shape (representing groups of vectors)


def get_smallest_possible(vals):
    """Returns smallest (fewest bits) possible DataType that can represent
    value. Prefers unsigned integers where possible."""
    vals = np.array(vals, dtype=np.float64)
    for v in vals:
        assert int(v) == v, "Error float value"

    for k in DataType.get_accumulator_dt_cands():
        dt = DataType[k]

        if dt in [DataType["BIPOLAR"], DataType["TERNARY"], DataType["FLOAT32"]]:
            # not currently supported
            continue

        if (dt.min() <= vals).all() and (vals <= dt.max()).all():
            return dt

    log.warning(
        """InferChannelwiseLinearLayer: Output values may not be
    representable with supported data types.
    Setting maximum width data type available.
    This will lead to errors if there are no constrains on the input
    """
    )

    if (0 <= vals).all():
        return DataType["UINT64"]
    else:
        return DataType["INT64"]


class ChannelwiseOp(HWCustomOp):
    """Abstraction layer for HW implementation of ChannelwiseOp."""

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {
            # channelwise "map" function to apply:
            # one of cmp_le, cmp_ge, add, mul
            "Func": ("s", False, "cmp_le", {"cmp_le", "cmp_ge", "add", "mul"}),
            "PE": ("i", True, 0),
            "NumChannels": ("i", True, 0),
            # string defining memory resource type for parameters
            "ram_style": ("s", False, "distributed", {"distributed", "block"}),
            # FINN DataTypes for inputs, weights, outputs
            "inputDataType": ("s", True, ""),
            "paramDataType": ("s", True, ""),
            "outputDataType": ("s", True, ""),
            # number of input vectors, examples:
            # [1] is a single vector (like a FC layer with batch=1)
            # [4] is four vectors (like a FC layer with batch=4)
            # [1, 4, 4] is four * four vectors (like a conv layer with batch=1)
            "numInputVectors": ("ints", False, [1]),
        }
        my_attrs.update(super().get_nodeattr_types())
        return my_attrs

    def calc_tmem(self):
        """Calculates and returns TMEM, the depth of the memory used
        to store the channelwise op parameters."""
        chn = self.get_nodeattr("NumChannels")
        pe = self.get_nodeattr("PE")
        return chn // pe

    def infer_node_datatype(self, model):
        node = self.onnx_node
        # check input datatype against property
        idt = model.get_tensor_datatype(node.input[0])

        exp_idt_name = self.get_nodeattr("inputDataType")
        if exp_idt_name != idt.name:
            func = self.get_nodeattr("Func")
            assert func in ["add", "mul"], "Bad input DataType for ChannelwiseOp layer"

            self.set_nodeattr("inputDataType", idt.name)
            # update the func in ['add','mul'] cases

            # get parameter ranges
            param = model.get_initializer(node.input[1])
            param_min = min(param.flatten())
            param_max = max(param.flatten())

            # set function and determine output data type
            if func == "add":
                out_min = idt.min() + param_min
                out_max = idt.max() + param_max
                odt = get_smallest_possible([out_min, out_max])
            elif func == "mul":
                possible_limits = []
                possible_limits += [idt.min() * param_min]
                possible_limits += [idt.min() * param_max]
                possible_limits += [idt.max() * param_min]
                possible_limits += [idt.max() * param_max]
                odt = get_smallest_possible(possible_limits)

            self.set_nodeattr("outputDataType", odt.name)

        # set output datatype from property
        odt = self.get_output_datatype()
        model.set_tensor_datatype(node.output[0], odt)

    def get_input_datatype(self, ind=0):
        """Returns FINN DataType of input."""
        if ind == 0:
            return DataType[self.get_nodeattr("inputDataType")]
        elif ind == 1:
            return DataType[self.get_nodeattr("paramDataType")]
        else:
            raise Exception("Undefined input ind for this layer type")

    def get_output_datatype(self, ind=0):
        """Returns FINN DataType of output."""
        return DataType[self.get_nodeattr("outputDataType")]

    def get_instream_width(self, ind=0):
        if ind == 0:
            i_bits = self.get_input_datatype().bitwidth()
            return i_bits * self.get_nodeattr("PE")
        elif ind == 1:
            # param input is not exposed so width can default to 0
            return 0
        else:
            raise Exception("Undefined input ind for this layer type")

    def get_outstream_width(self, ind=0):
        o_bits = self.get_output_datatype().bitwidth()
        return o_bits * self.get_nodeattr("PE")

    def get_folded_input_shape(self, ind=0):
        if ind == 0:
            ich = self.get_nodeattr("NumChannels")
            pe = self.get_nodeattr("PE")
            fold = ich // pe
            vecs = list(self.get_nodeattr("numInputVectors"))
            folded_input_shape = tuple(vecs + [fold, pe])
            return folded_input_shape
        elif ind == 1:
            return self.get_normal_input_shape(ind)
        else:
            raise Exception("Undefined input ind for this layer type")

    def get_folded_output_shape(self, ind=0):
        # same shape as input
        return self.get_folded_input_shape()

    def get_normal_input_shape(self, ind=0):
        ich = self.get_nodeattr("NumChannels")
        if ind == 0:
            vecs = list(self.get_nodeattr("numInputVectors"))
            normal_input_shape = tuple(vecs + [ich])
        elif ind == 1:
            normal_input_shape = tuple(
                [ich],
            )
        else:
            raise Exception("Undefined input ind for this layer type")
        return normal_input_shape

    def get_normal_output_shape(self, ind=0):
        # same shape as input
        return self.get_normal_input_shape()

    def get_exp_cycles(self):
        # Channels/PE * batch size * fmdim * fmdim
        return np.prod(self.get_folded_output_shape()[:-1])

    def execute_node(self, context, graph):
        # create a standard onnx node to help calculate the result
        # depending on Func node attribute either a Mul or an Add node
        node = self.onnx_node
        func = self.get_nodeattr("Func")
        inp_values = context[node.input[0]]
        param_values = context[node.input[1]]
        oshape = context[node.output[0]].shape
        ishape = inp_values.shape
        pshape = param_values.shape
        inp = helper.make_tensor_value_info(node.input[0], TensorProto.FLOAT, ishape)
        param = helper.make_tensor_value_info(node.input[1], TensorProto.FLOAT, pshape)
        outp = helper.make_tensor_value_info(node.output[0], TensorProto.FLOAT, oshape)
        node_func = helper.make_node(
            func.capitalize(),
            inputs=node.input,
            outputs=[node.output[0]],
        )
        graph_func = helper.make_graph(
            nodes=[node_func],
            name="single-add-exec",
            inputs=[inp, param],
            outputs=[outp],
        )

        opset_version = self.onnx_opset_version
        opset_imports = [helper.make_opsetid("", opset_version)]
        onnx_kwargs = {"opset_imports": opset_imports}
        model_func = qonnx_make_model(graph_func, **onnx_kwargs)
        idict = {node.input[0]: inp_values, node.input[1]: param_values}
        sess = rt.InferenceSession(model_func.SerializeToString())
        result = sess.run(None, idict)
        context[node.output[0]] = np.asarray(result, dtype=np.float32).reshape(oshape)
